from dm.domain.entities import Scope, ActionType, LogMode
from dm.domain.entities.transfer import Status

UUID_pattern = "^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$"

login_post = {
    "type": "object",
    "properties": {
        "username": {"type": "string"},
        "password": {"type": "string"}

    },
    "required": ["username", "password"],
    "additionalProperties": False
}

schema_healthcheck = {
    "type": "object",
    "properties": {
        "action": {"type": "string",
                   "pattern": "^(reboot|stop|software)"},
    },
    "required": ["action"],
    "additionalProperties": False
}

locker_prevent_post = {
    "type": "object",
    "properties": {
        "scope": {"type": "string",
                  "pattern": f"^({'|'.join([s.name for s in Scope])})$"},
        "applicant": {"type": ["object", "array", "string"]},
        "datemark": {"type": "string"},
    },
    "required": ["scope", "applicant", "datemark"],
    "additionalProperties": False
}

locker_unlock_lock_post = {
    "type": "object",
    "properties": {
        "scope": {"type": "string",
                  "pattern": f"^({'|'.join([s.name for s in Scope])})$"},
        "applicant": {"type": ["object", "array", "string"]},
        # "force": {"type": "boolean"},
    },
    "required": ["scope", "applicant"],
    "additionalProperties": False
}


send_post = {
    "type": "object",
    "properties": {
        "software_id": {"type": "string",
                        "pattern": UUID_pattern},
        "software": {"type": "string"},
        "version": {"type": "string"},
        "file": {"type": "string"},
        "dest_server_id": {"type": "string",
                           "pattern": UUID_pattern},
        "dest_path": {"type": "string"},
        "chunk_size": {"type": "integer",  # size in MB
                       "minimum": 1,
                       "maximum": 1 * 1024,
                       # "multipleOf": 1024
                       },
        "max_senders": {"type": "integer",
                        "minimum": 1},
        "background": {"type": "boolean"},
        "force": {"type": "boolean"},
        "include_transfer_data": {"type": "boolean"},
    },
    "oneOf": [{"required": ["software_id", "dest_server_id"]},
              {"required": ["software", "version", "dest_server_id"]},
              {"required": ["file", "dest_server_id"]}],
    "additionalProperties": False
}

software_post = {
    "type": "object",
    "properties": {
        "name": {"type": "string"},
        "version": {"type": "string"},
        "family": {"type": "string"},
        "file": {"type": "string"}
    },
    "required": ["name", "version", "file"],
    "additionalProperties": False
}

software_patch = {
    "type": "object",
    "properties": {
        "server_id": {"type": "string",
                      "pattern": UUID_pattern},
        "path": {"type": "string"},
    },
    "required": ["server_id", "path"],
    "additionalProperties": False
}

software_servers_put = {
    "type": "array",
    "items": {
        "type": "object",
        "properties": {
            "server_id": {"type": "string",
                          "pattern": UUID_pattern},
            "file": {"type": "string"}
        },
        "required": ["server_id", "path"]
    },
    "additionalProperties": False
}

action_type_pattern = f"^({'|'.join([at.name for at in ActionType if at.name != 'NATIVE'])})$"
action_template_post = {
    "type": "object",
    "properties": {
        "name": {"type": "string"},
        "action_type": {"type": "string",
                        "pattern": action_type_pattern},
        "code": {"type": "string"},
        "parameters": {"type": "object"},
        "expected_stdout": {"type": "string"},
        "expected_stderr": {"type": "string"},
        "expected_rc": {"type": "integer"},
        "system_kwargs": {"type": "object"},
        "pre_process": {"type": "string"},
        "post_process": {"type": "string"},
    },
    "if": {
        "properties": {"action_type": {"const": "SEND"}}
    },
    "then": {
        "required": ["name", "action_type"],
    },
    "else": {
        "properties": {"code": {"type": "string"}},
        "required": ["name", "action_type", "code"],
    },
    "required": ["name", "action_type", "code"],
    "additionalProperties": False
}

action_template_patch = {
    "type": "object",
    "properties": {
        "action_type": {"type": "string",
                        "pattern": f"^({'|'.join([at.name for at in ActionType])})$"},
        "code": {"type": "string"},
        "parameters": {"type": "object"},
        "expected_stdout": {"type": "string"},
        "expected_stderr": {"type": "string"},
        "expected_rc": {"type": "integer"},
        "system_kwargs": {"type": "object"},
        "pre_process": {"type": "string"},
        "post_process": {"type": "string"},
    },
    "additionalProperties": False
}

_step_post = {
    "type": "object",
    "properties": {
        "id": {"anyOf": [{"type": "string"},
                         {"type": "integer", "minimum": 1}]},
        "orchestration_id": {"type": "string",
                             "pattern": UUID_pattern},
        "undo": {"type": "boolean"},
        "name": {"type": "string"},
        "stop_on_error": {"type": "boolean"},
        "stop_undo_on_error": {"type": "boolean"},
        "action_template_id": {"type": "string",
                               "pattern": UUID_pattern},
        "action_type": {"type": "string",
                        "pattern": action_type_pattern},
        "undo_on_error": {"type": "boolean"},
        "code": {"type": "string"},
        "parameters": {"type": "object"},
        "expected_stdout": {"type": "string"},
        "expected_stderr": {"type": "string"},
        "expected_rc": {"type": "integer"},
        "system_kwargs": {"type": "object"},
        "pre_process": {"type": "string"},
        "post_process": {"type": "string"},
        "parent_step_ids": {"type": "array",
                            "items": {"anyOf": [{"type": "string", "pattern": UUID_pattern},
                                                {"type": "integer", "minimum": 1}]
                                      }
                            },
        "target": {"anyOf": [{"type": "string"},
                             {"type": "array", "items": {"type": "string"}}]}
    },
    "oneOf": [{"required": ["orchestration_id", "undo", "action_template_id"]},
              {"required": ["orchestration_id", "undo", "action_type"]}],
    "additionalProperties": False
}

step_post = {
    "anyOf": [
        {"type": "array",
         "items": _step_post},
        _step_post
    ]
}

step_put = {
    "type": "object",
    "properties": {
        "undo": {"type": "boolean"},
        "stop_on_error": {"type": "boolean"},
        "stop_undo_on_error": {"type": "boolean"},
        "undo_on_error": {"type": "boolean"},
        "action_template_id": {"type": "string",
                               "pattern": UUID_pattern},
        "action_type": {"type": "string",
                        "pattern": action_type_pattern},

        "code": {"type": "string"},
        "parameters": {"type": "object"},
        "expected_stdout": {"type": "string"},
        "expected_stderr": {"type": "string"},
        "expected_rc": {"type": "integer"},
        "system_kwargs": {"type": "object"},
        "pre_process": {"type": "string"},
        "post_process": {"type": "string"},
        "name": {"type": "string"},
        "parent_step_ids": {"type": "array",
                            "items": {"anyOf": [{"type": "string", "pattern": UUID_pattern},
                                                {"type": "integer", "minimum": 1}]
                                      }
                            },
        "child_step_ids": {"type": "array",
                           "items": {"anyOf": [{"type": "string", "pattern": UUID_pattern},
                                               {"type": "integer", "minimum": 1}]
                                     }
                           },
        "target": {"anyOf": [{"type": "string"},
                             {"type": "array", "items": {"type": "string"}}]}
    },
    "oneOf": [{"required": ["undo", "action_template_id"]},
              {"required": ["undo", "action_type"]}],
    "additionalProperties": False
}

step_patch = {
    "type": "object",
    "properties": {
        "undo": {"type": "boolean"},
        "stop_on_error": {"type": "boolean"},
        "stop_undo_on_error": {"type": "boolean"},
        "action_template_id": {"type": "string",
                               "pattern": UUID_pattern},
        "action_type": {"type": "string",
                        "pattern": action_type_pattern},
        "undo_on_error": {"type": "boolean"},
        "code": {"type": "string"},
        "parameters": {"type": "object"},
        "expected_stdout": {"type": "string"},
        "expected_stderr": {"type": "string"},
        "expected_rc": {"type": "integer"},
        "system_kwargs": {"type": "object"},
        "pre_process": {"type": "string"},
        "post_process": {"type": "string"},
        "name": {"type": "string"},
        "parent_step_ids": {"type": "array",
                            "items": {"anyOf": [{"type": "string", "pattern": UUID_pattern},
                                                {"type": "integer", "minimum": 1}]
                                      }
                            },
        "child_step_ids": {"type": "array",
                           "items": {"anyOf": [{"type": "string", "pattern": UUID_pattern},
                                               {"type": "integer", "minimum": 1}]
                                     }
                           },
        "target": {"anyOf": [{"type": "string"},
                             {"type": "array", "items": {"type": "string"}}]}
    },
    "additionalProperties": False
}

step_parents = {
    "type": "object",
    "properties": {
        "parent_step_ids": {"type": "array", "items": {"type": "string", "pattern": UUID_pattern}},
        "additionalProperties": False
    }
}

step_children = {
    "type": "object",
    "properties": {
        "children_step_ids": {"type": "array", "items": {"type": "string", "pattern": UUID_pattern}},
        "additionalProperties": False
    }
}

routes_patch = {
    "type": "object",
    "properties": {
        "discover_new_neighbours": {"type": "boolean"},
        "check_current_neighbours": {"type": "boolean"},
        "server_id": {"type": "string",
                      "pattern": UUID_pattern},
        "server_list": {"type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "id": {"type": "string", "pattern": UUID_pattern},
                                "gateway": {"anyOf": [
                                    {"type": "string",
                                     "pattern": UUID_pattern},
                                    {"type": "null"}
                                ]},
                                "cost": {"anyOf": [
                                    {"type": "integer",
                                     "minimum": 0},
                                    {"type": "null"}
                                ]}
                            },
                            "required": ["id", "gateway", "cost"]
                        }
                        }
    },
    "additionalProperties": False
}

routes_post = {
    "type": "object",
    "properties": {
        "discover_new_neighbours": {"type": "boolean"},
        "check_current_neighbours": {"type": "boolean"},
        "retries": {"type": "integer", "minimum": 1},
        "timeout": {"type": "number", "minimum": 0}
    },
    "additionalProperties": False
}

orchestration_post = {
    "type": "object",
    "properties": {
        "name": {"type": "string"},
        "description": {"type": "string"},
        "stop_on_error": {"type": "boolean"},
        "stop_undo_on_error": {"type": "boolean"},
        "undo_on_error": {"type": "boolean"}
    },
    "required": ["name"],
    "additionalProperties": False
}

orchestration_step = {
    "type": "object",
    "properties": {
        "id": {"anyOf": [{"type": "string"},
                         {"type": "integer", "minimum": 1}]},
        "undo": {"type": "boolean"},
        "name": {"type": "string"},
        "stop_on_error": {"type": "boolean"},
        "stop_undo_on_error": {"type": "boolean"},
        "action_template_id": {"type": "string",
                               "pattern": UUID_pattern},
        "action_type": {"type": "string",
                        "pattern": action_type_pattern},
        "undo_on_error": {"type": "boolean"},
        "code": {"type": "string"},
        "parameters": {"type": "object"},
        "expected_stdout": {"type": "string"},
        "expected_stderr": {"type": "string"},
        "expected_rc": {"type": "integer"},
        "system_kwargs": {"type": "object"},
        "pre_process": {"type": "string"},
        "post_process": {"type": "string"},
        "parent_step_ids": {"type": "array",
                            "items": {"anyOf": [{"type": "string", "pattern": UUID_pattern},
                                                {"type": "integer", "minimum": 1}]
                                      }
                            },
        "target": {"anyOf": [{"type": "string"},
                             {"type": "array", "items": {"type": "string"}}]}
    },
    "oneOf": [{"required": ["undo", "action_template_id"]},
              {"required": ["undo", "action_type"]}],
    "additionalProperties": False
}


orchestration_full = {
    "type": "object",
    "properties": {
        "name": {"type": "string"},
        # "version": {"type": "integer", "minimum": 1},
        "description": {"type": "string"},
        "stop_on_error": {"type": "boolean"},
        "stop_undo_on_error": {"type": "boolean"},
        "undo_on_error": {"type": "boolean"},
        "steps": {"type": "array",
                  "items": orchestration_step}
    },
    "required": ["name", "steps"],
    "additionalProperties": False
}

orchestration_patch = {
    "type": "object",
    "properties": {
        "name": {"type": "string"},
        "version": {"type": "integer", "minimum": 1},
        "description": {"type": "string"},
        "stop_on_error": {"type": "boolean"},
        "stop_undo_on_error": {"type": "boolean"},
        "undo_on_error": {"type": "boolean"}
    },
    "additionalProperties": False
}


launch_orchestration_post = {
    "type": "object",
    "properties": {
        'hosts': {"anyOf": [{"type": "string"},
                            {"type": "array",
                             "items": {"type": "string"},
                             "minItems": 1},
                            {"type": "object",
                             "patternProperties": {
                                 ".*": {"anyOf": [{"type": "string"},
                                                  {"type": "array",
                                                   "items": {"type": "string"},
                                                   "minItems": 1
                                                   },
                                                  ]
                                        },
                             },
                             }]
                  },
        "params": {"type": "object"},
        "background": {"type": "boolean"}
    },
    "required": ["hosts"],
    "additionalProperties": False,
}



transfers_post = {
    "type": "object",
    "properties": {
        "software_id": {"type": "string",
                        "pattern": UUID_pattern},
        "filename": {"type": "string"},         # specify filename if you want to send a file instead of a software
        "size": {"type": "integer"},
        "checksum": {"type": "string"},
        "dest_path": {"type": "string"},        # if not specified DM_SOFTWARE_REPO is used
        "num_chunks": {"type": "integer",
                       "minimum": 0},
        "cancel_pending": {"type": "boolean"},  # cancels pending transfers from the same file in the same folder
        "force": {"type": "boolean"},           # forces to transfer file even if it exists in the destination

    },
    "required": ['num_chunks'],
    "additionalProperties": False
}

transfer_post = {
    "type": "object",
    "properties": {
        "chunk": {"type": "integer",
                  "minimum": 0},
        "content": {"type": "string"},
    },
    "required": ["chunk", "content"],
    "additionalProperties": False
}

transfer_patch = {
    "type": "object",
    "properties": {
        "status": {"type": "string",
                   "pattern": f"^({'|'.join([s.name for s in Status])})$"},
    },
    "required": ["status"],
    "additionalProperties": False
}

logs_post = {
    "type": "object",
    "properties": {
        "src_server_id": {"type": "string",
                          "pattern": UUID_pattern},
        "target": {"type": "string"},
        "include": {"type": "string"},
        "exclude": {"type": "string"},
        "dst_server_id": {"type": "string",
                          "pattern": UUID_pattern},
        "dest_folder": {"type": "string"},
        "recursive": {"type": "boolean"},
        "mode": {"type": "string",
                 "pattern": f"^({'|'.join([m.name for m in LogMode])})$"}
    },
    "additionalProperties": False,
    "required": ["src_server_id", "target", "dst_server_id"],
}

log_patch = {
    "type": "object",
    "properties": {
        "include": {"type": "string",
                    "format": "regex"},
        "exclude": {"type": "string",
                    "format": "regex"},
        "dest_folder": {"type": "string"},
        "recursive": {"type": "boolean"},
        "mode": {"type": "string",
                 "pattern": f"^({'|'.join([m.name for m in LogMode])})$"}
    },
    "additionalProperties": False
}

log_post = {
    "type": "object",
    "properties": {
        "file": {"type": "string"},
        "data": {"type": "string"},
    },
    "required": ["data"],
    "additionalProperties": False
}

users_post = {
    "type": "object",
    "properties": {
        "user": {"type": "string"},
        "password": {"type": "string"},
        "email": {"type": "string",
                  "format": "email"}
    },
    "required": ["user", "password", "email"],
    "additionalProperties": False
}

user_patch = {
    "type": "object",
    "properties": {
        "email": {"type": "string",
                  "format": "email"},
        "active": {"type": "boolean"}
    },
    "additionalProperties": False
}

launch_command_post = {
    "type": "object",
    "properties": {
        "command": {"type": "string"},
        "hosts": {"anyOf": [{"type": "string"},
                            {"type": "array",
                             "items": {"type": "string"}},
                            ]
                  },
        "timeout": {"type": "integer",
                    "minimum": 1},
        "input": {"type": "string"},
    },
    "required": ["command"],
    "additionalProperties": False,
}

server_post = {
    "type": "object",
    "properties": {
        # "gates": {"type": "object",
        #           "properties": {
        #               "dns_or_ip": {"type": "string"},
        #               "port": {"type": "integer",
        #                        "minimum": 1,
        #                        "maximum": 65535},
        #               "hidden": {"type": "boolean"},
        #               "required": ["dns_or_ip", "port"],
        #               "additionalProperties": False,
        #           }},
        "granules": {"type": "array",
                     "items": {"type": "string"}}
    },
    "additionalProperties": False,
    "required": ["granules"]
}

server_patch = {
    "type": "object",
    "properties": {
        "gates": {"type": "array",
                  "items": {"type": "object",
                            "properties": {
                                "dns_or_ip": {"type": "string"},
                                "port": {"type": "integer",
                                         "minimum": 1,
                                         "maximum": 65535},
                                "hidden": {"type": "boolean"},
                                "required": ["dns_or_ip", "port"],
                                "additionalProperties": False,
                            }}
                  },
        "granules": {"type": "array",
                     "items": {"type": "string"}}
    },
    "additionalProperties": False,
}



from unittest import TestCase, mock

from prompt_toolkit.completion import CompleteEvent, WordCompleter
from prompt_toolkit.document import Document

from dm.dshell.completer import ResourceCompleter, DshellCompleter
from dm.web.network import Response


class TestResourceCompleter(TestCase):
    default_response = Response(msg=[
        {
            "last_modified_at": "20200616.113822.505932+0200",
            "id": "00000000-1111-0000-0000-000000000000",
            "name": "dev1",
            "granules": []
        },
        {
            "last_modified_at": "20200616.113955.580271+0200",
            "id": "00000000-2222-0000-0000-000000000000",
            "name": "dev2",
            "granules": []
        },
        {
            "last_modified_at": "20200616.113955.580271+0200",
            "id": "00000000-3333-0000-0000-000000000000",
            "name": "node3",
            "granules": []
        }
    ], code=200)

    @mock.patch('dm.dshell.completer.request')
    def test_get_completions(self, mock_request):
        mock_request.return_value = self.default_response

        completer = ResourceCompleter('servers', 'name')
        # Static list on empty input.
        completions = completer.get_completions(Document(""), CompleteEvent())
        self.assertListEqual(["dev1", "dev2", "node3"], [c.text for c in completions])

        completions = completer.get_completions(Document("d"), CompleteEvent())
        self.assertListEqual(["dev1", "dev2"], [c.text for c in completions])

    @mock.patch('dm.dshell.completer.request')
    def test_filter(self, mock_request):
        mock_request.return_value = self.default_response

        completer = ResourceCompleter('servers', 'name', match_middle=True)

        completions = completer.get_completions(Document("3"), CompleteEvent())
        self.assertListEqual(["node3"], [c.text for c in completions])

        completer = ResourceCompleter('servers', 'name', ignore_case=True)

        completions = completer.get_completions(Document("D"), CompleteEvent())
        self.assertListEqual(["dev1", "dev2"], [c.text for c in completions])


class TestDshellCompleter(TestCase):

    @mock.patch('dm.dshell.completer.request')
    def test_get_completions(self, mock_request):
        mock_request.return_value = Response(msg=[
            {
                "last_modified_at": "20200616.113822.505932+0200",
                "id": "01",
                "name": "dev1",
                "granules": []
            },
            {
                "last_modified_at": "20200616.113955.580271+0200",
                "id": "02",
                "name": "dev2",
                "granules": []
            },
            {
                "last_modified_at": "20200616.113955.580271+0200",
                "id": "13",
                "name": "node3",
                "granules": []
            }
        ], code=200)

        data = {
            'status': [{'argument': 'node', 'completer': ResourceCompleter('servers', 'name')}],
            'server': {
                'list': [{'argument': '--json', 'action': 'store_true', 'required': False},
                         [{'argument': '--like'},
                          {'argument': '--name', 'completer': ResourceCompleter('servers', 'name')},
                          {'argument': '--id', 'completer': ResourceCompleter('servers', 'id')},
                          {'argument': '--last', 'action': 'store', 'type': int}]
                         ],
                'show': [{'argument': '--json', 'action': 'store_true', 'required': False},
                         {'argument': 'node', 'completer': ResourceCompleter('servers', 'name')}]},
            'nargs': [{'argument': 'compiler', 'completer': WordCompleter(['python', 'ruby', 'c++']), 'nargs': 2},
                      {'argument': 'so', 'completer': WordCompleter(['windows', 'linux', 'mac'])},
                      {'argument': '--foo', 'nargs': '?', 'completer': WordCompleter(['aaa', 'bbb'])},
                      {'argument': '--bar', 'nargs': '*', 'completer': WordCompleter(['x1', 'x2', 'x3'])}]
        }

        completer = DshellCompleter.from_nested_dict(data)
        # Static list on empty input.
        completions = completer.get_completions(Document(""), CompleteEvent())
        self.assertListEqual(["status", "server", "nargs"], [c.text for c in completions])

        # mid word
        completions = completer.get_completions(Document("st"), CompleteEvent())
        self.assertListEqual(["status"], [c.text for c in completions])

        # completed word
        completions = completer.get_completions(Document("status"), CompleteEvent())
        self.assertListEqual([], [c.text for c in completions])

        completions = completer.get_completions(Document("status "), CompleteEvent())
        self.assertListEqual(["dev1", "dev2", "node3"], [c.text for c in completions])

        completions = completer.get_completions(Document("server "), CompleteEvent())
        self.assertListEqual(["list", "show"], [c.text for c in completions])

        completions = completer.get_completions(Document("server l"), CompleteEvent())
        self.assertListEqual(["list"], [c.text for c in completions])

        completions = completer.get_completions(Document("server list "), CompleteEvent())
        self.assertListEqual(["--json", "--like", "--name", "--id", "--last"], [c.text for c in completions])

        completions = completer.get_completions(Document("server list --j"), CompleteEvent())
        self.assertListEqual(["--json"], [c.text for c in completions])

        completions = completer.get_completions(Document("server list --json "), CompleteEvent())
        self.assertListEqual(["--json", "--like", "--name", "--id", "--last"], [c.text for c in completions])

        completions = completer.get_completions(Document("server list --json --id "), CompleteEvent())
        self.assertListEqual(["01", "02", "13"], [c.text for c in completions])

        completions = completer.get_completions(Document("server list --json --id 0"), CompleteEvent())
        self.assertListEqual(["01", "02"], [c.text for c in completions])

        completions = completer.get_completions(Document("server list --json --id 01 "), CompleteEvent())
        self.assertListEqual(["--json", "--like", "--name", "--id", "--last"], [c.text for c in completions])

        completions = completer.get_completions(Document("server show "), CompleteEvent())
        self.assertListEqual(["--json", "dev1", "dev2", "node3"], [c.text for c in completions])

        completions = completer.get_completions(Document("nargs "), CompleteEvent())
        self.assertListEqual(['--foo', '--bar', 'python', 'ruby', 'c++'], [c.text for c in completions])

        completions = completer.get_completions(Document("nargs --foo "), CompleteEvent())
        self.assertListEqual(['aaa', 'bbb', '--foo', '--bar', 'python', 'ruby', 'c++'], [c.text for c in completions])

        completions = completer.get_completions(Document("nargs --foo s "), CompleteEvent())
        self.assertListEqual(['--foo', '--bar', 'python', 'ruby', 'c++'], [c.text for c in completions])

        completions = completer.get_completions(Document("nargs --bar "), CompleteEvent())
        self.assertListEqual(['x1', 'x2', 'x3', '--foo', '--bar', 'python', 'ruby', 'c++'],
                             [c.text for c in completions])

        completions = completer.get_completions(Document("nargs --bar x"), CompleteEvent())
        self.assertListEqual(['x1', 'x2', 'x3'], [c.text for c in completions])

        completions = completer.get_completions(Document("nargs --bar x1 "), CompleteEvent())
        self.assertListEqual(['x1', 'x2', 'x3', '--foo', '--bar', 'python', 'ruby', 'c++'],
                             [c.text for c in completions])



from unittest import TestCase

from flask_jwt_extended import create_access_token

from dm.domain.entities import User
from dm.network.auth import HTTPBearerAuth
from dm.web import create_app, db


class Test(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.auth = HTTPBearerAuth(create_access_token('00000000-0000-0000-0000-000000000001'))
        db.create_all()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_verify_password(self):
        u = User(user='root', email=None, groups=['adm', 'dev'])
        u.set_password('test')
        self.assertIsNotNone(u._password)

        self.assertTrue(u.verify_password('test'))

    def test_to_from(self):
        u = User(user='root', email=None, groups=['adm', 'dev'])
        db.session.add(u)

        u_json = u.to_json(password=True)

        smashed = u.from_json(u_json)

        self.assertEqual(u, smashed)
        self.assertEqual(u.id, smashed.id)
        self.assertEqual(u.active, smashed.active)
        self.assertEqual(u.created_at, smashed.created_at)
        self.assertEqual(u.email, smashed.email)
        self.assertEqual(u.groups, smashed.groups)

        db.session.delete(u)

        smashed = u.from_json(u_json)

        self.assertEqual(u.id, smashed.id)
        self.assertEqual(u.active, smashed.active)
        self.assertEqual(u.created_at, smashed.created_at)
        self.assertEqual(u.email, smashed.email)
        self.assertEqual(u.groups, smashed.groups)

    def test_to_from_no_password(self):
        u = User(user='root', email=None, groups=['adm', 'dev'])
        db.session.add(u)

        u_json = u.to_json()

        smashed = User.from_json(u_json)

        self.assertEqual(u, smashed)
        self.assertEqual(u.id, smashed.id)
        self.assertEqual(u.active, smashed.active)
        self.assertEqual(u.created_at, smashed.created_at)
        self.assertEqual(u.email, smashed.email)
        self.assertEqual(u.groups, smashed.groups)

        db.session.delete(u)

        smashed = User.from_json(u_json)
        db.session.add(smashed)
        db.session.flush()

        self.assertEqual(u.id, smashed.id)
        self.assertEqual(u.active, smashed.active)
        self.assertEqual(u.created_at, smashed.created_at)
        self.assertEqual(u.email, smashed.email)
        self.assertEqual(u.groups, smashed.groups)





from unittest import TestCase

from flask_jwt_extended import create_access_token

from dm.domain.entities import Log, Server
from dm.domain.entities.log import Mode
from dm.network.auth import HTTPBearerAuth
from dm.web import create_app, db


class TestLog(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.auth = HTTPBearerAuth(create_access_token('00000000-0000-0000-0000-000000000001'))
        db.create_all()
        self.src = Server('source', port=5000)
        self.dst = Server('destination', port=5000)

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_to_from_json(self):
        l = Log(source_server=self.src, target='/home/dimensigon/dimensigon/dm.log', destination_server=self.dst,
                dest_folder='/home/dimensigon/dimensigon-node3', id='11111111-2222-3333-4444-111111110001')

        self.src.id = '11111111-2222-3333-4444-555555550001'
        self.dst.id = '11111111-2222-3333-4444-555555550002'
        db.session.add(self.src)
        db.session.add(self.dst)

        l_json = l.to_json()
        self.assertDictEqual(
            dict(id='11111111-2222-3333-4444-111111110001',
                 src_server_id='11111111-2222-3333-4444-555555550001', target='/home/dimensigon/dimensigon/dm.log',
                 include=None,
                 exclude=None, dst_server_id='11111111-2222-3333-4444-555555550002',
                 dest_folder='/home/dimensigon/dimensigon-node3',
                 recursive=False,
                 mode='FOLDER',
                 deleted=False,
                 _old_target=None), l_json)

        smashed = Log.from_json(l_json)
        self.assertEqual(smashed.source_server, self.src)
        self.assertEqual(smashed.destination_server, self.dst)
        self.assertIsNone(smashed.include)
        self.assertIsNone(smashed.exclude)
        self.assertEqual(smashed.dest_folder, '/home/dimensigon/dimensigon-node3')
        self.assertEqual(smashed.target, '/home/dimensigon/dimensigon/dm.log')
        self.assertFalse(smashed.recursive)
        self.assertFalse(smashed.deleted)
        self.assertEqual(Mode.FOLDER, smashed.mode)
        self.assertIsNone(smashed._old_target)

        l_json = l.to_json(human=True)
        self.assertDictEqual(
            dict(id='11111111-2222-3333-4444-111111110001',
                 src_server=self.src.name, target='/home/dimensigon/dimensigon/dm.log',
                 include=None,
                 exclude=None, dst_server=self.dst.name,
                 dest_folder='/home/dimensigon/dimensigon-node3',
                 recursive=False,
                 mode='FOLDER',
                 deleted=False,
                 _old_target=None), l_json)

        l_json = l.to_json(human=True, delete_data=False)
        self.assertDictEqual(
            dict(id='11111111-2222-3333-4444-111111110001',
                 src_server=self.src.name, target='/home/dimensigon/dimensigon/dm.log',
                 include=None,
                 exclude=None, dst_server=self.dst.name,
                 dest_folder='/home/dimensigon/dimensigon-node3',
                 recursive=False,
                 mode='FOLDER'
                 ), l_json)

    def test_delete(self):
        l = Log(source_server=self.src, target='/home/dimensigon/dimensigon/dm.log', destination_server=self.dst,
                dest_folder='/home/dimensigon/dimensigon-node3', id='11111111-2222-3333-4444-111111110001')

        l.delete()

        self.assertEqual('/home/dimensigon/dimensigon/dm.log', l._old_target)
        self.assertNotEqual('/home/dimensigon/dimensigon/dm.log', l.target)
        self.assertTrue(l.deleted)

        db.session.add(l)
        db.session.commit()

        log_list = l.query.all()
        self.assertEqual(0, len(log_list))
        l = db.session.query(Log).filter(id='11111111-2222-3333-4444-111111110001').one()


import logging
import os
import re
import urllib

import requests
from prompt_toolkit import prompt

from dm import defaults
from dm.dshell import environ
from dm.dshell.view_path_mapping import view_path_map
from dm.network.auth import HTTPBearerAuth
from dm.utils.helpers import is_iterable_not_string
from dm.web.network import Response

logger = logging.getLogger('dshell.network')

_access_token = None
_refresh_token = os.environ.get('DM_REFRESH_TOKEN')
_username = None


def bootstrap_auth(username=None, password=None, refresh_token=None):
    global _username, _refresh_token

    if username:
        _username = username
    if refresh_token:
        _refresh_token = refresh_token

    if not exists_refresh_token():
        login(username=username, password=password)
    else:
        try:
            refresh_username = refresh_access_token()
        except Exception:
            raise
        else:
            if refresh_username:
                if _username is None:
                    _username = refresh_username
                elif _username != refresh_username:
                    raise ValueError(f"user '{_username}' does not correspond with the user from the token")


def login(username=None, password=None):
    global _access_token, _refresh_token, _username

    if not username:
        if not _username:
            username = prompt("Username: ")
        else:
            username = _username
    if not password:
        password = prompt("Password: ", is_password=True)

    resp = requests.post(generate_url('root.login', {}), json={'username': username, 'password': password},
                         verify=False)
    resp.raise_for_status()

    _username = username
    _access_token = resp.json()['access_token']
    _refresh_token = resp.json()['refresh_token']


def exists_refresh_token():
    return bool(_refresh_token)


def refresh_access_token():
    global _refresh_token, _access_token, _username
    url = f"{environ.get('SCHEME')}://{environ.get('SERVER')}:{environ.get('PORT')}/refresh"

    if _refresh_token is None:
        raise ValueError('empty refresh token. Login first')
    auth = HTTPBearerAuth(_refresh_token)
    resp = requests.post(url, auth=auth, verify=False, timeout=10)
    resp.raise_for_status()
    _access_token = resp.json().get('access_token', None)
    return resp.json().get('username', None)


def request(method, url, session=None, token_refreshed=False, **kwargs) -> Response:
    global _access_token
    exception = None
    content = None
    status = None
    json_data = None
    headers = {}

    if not session:
        _session = requests.session()
    else:
        _session = session

    raise_on_error = kwargs.pop('raise_on_error', False)

    func = getattr(_session, method.lower())

    if _access_token is None:
        try:
            refresh_access_token()
        except requests.exceptions.ConnectionError as e:
            return Response(exception=ValueError(f"Unable to contact with {environ.get('SCHEME')}://"
                                                 f"{environ.get('SERVER')}:{environ.get('PORT')}/"),
                            url=url)
        except Exception as e:
            return Response(exception=e, url=url)
    kwargs['auth'] = HTTPBearerAuth(_access_token)

    if 'headers' not in kwargs:
        kwargs['headers'] = {}
    kwargs['headers'].update({'D-Securizer': 'plain'})

    kwargs['verify'] = False

    logger.debug(f"{method.upper()} {url}\n{kwargs}")

    resp = None
    try:
        resp: requests.Response = func(url, **kwargs)
    except (requests.Timeout,) as e:
        timeout = kwargs.get('timeout', None)
        if isinstance(timeout, tuple):
            timeout = timeout[0] + timeout[1]
        exception = TimeoutError(f"Socket timeout reached while trying to connect to {url} "
                                 f"for {timeout} seconds")
    except Exception as e:
        exception = e
    finally:
        if session is None and not (
                getattr(resp, 'status_code', None) == 401 and resp.json().get('msg') == 'Token has expired'):
            _session.close()
    if exception is None:
        status = resp.status_code
        headers = resp.headers
        if status == 401 and not token_refreshed:
            json_data = resp.json()
            if json_data.get('msg') == 'Token has expired':
                try:
                    _access_token = refresh_access_token()
                except requests.exceptions.ConnectionError as e:
                    return Response(exception=ValueError(f"Unable to contact with {environ.get('SCHEME')}://"
                                                         f"{environ.get('SERVER')}:{environ.get('PORT')}/"),
                                    url=url)
                resp = request(method, url, session=_session, token_refreshed=True, **kwargs)
                if not session:
                    _session.close()
                return resp
        try:
            json_data = resp.json()
        except (ValueError,):
            content = resp.text

        if json_data is not None:
            # try:
            #     content = unpack_msg(json_data)
            # except NotValidMessage:
            #     content = json_data
            content = json_data
    else:
        if raise_on_error:
            raise exception

    return Response(msg=content, code=status, exception=exception, url=url, headers=headers)


def get_parameters_from_path(view):
    path = view_path_map[view]
    match_iterator = re.finditer('\<([\w_]+)\>', path)
    for match in match_iterator:
        yield match.groups()[0]


def _replace_path_args(path, args):
    match_iterator = re.finditer('\<([\w_]+)\>', path)
    replaced_path = path
    for match in match_iterator:
        text = match.groups()[0]
        assert text in args
        replaced_path = replaced_path[:match.start()] + args.pop(text) + replaced_path[match.end():]
    params = []
    if args:
        for k, v in args.items():
            if is_iterable_not_string(v):
                for vv in v:
                    if vv is not None:
                        params.append(f"{k}={urllib.parse.quote_plus(vv)}")
            else:
                if v is not None:
                    params.append(f"{k}={urllib.parse.quote_plus(v)}")
    return replaced_path + '?' + '&'.join(params)


def generate_url(view, view_data):
    path = view_path_map[view]
    path = _replace_path_args(path, view_data or {})
    if environ.get('SERVER') is None:
        raise ValueError('No SERVER specified.')
    return f"{environ.get('SCHEME', 'https') or 'https'}://{environ.get('SERVER')}:{environ.get('PORT', defaults.DEFAULT_PORT) or defaults.DEFAULT_PORT}{path}"


def get(view, view_data=None, **kwargs) -> Response:
    return request('get', generate_url(view, view_data), **kwargs)


def post(view, view_data=None, **kwargs) -> Response:
    return request('post', generate_url(view, view_data), **kwargs)


def put(view, view_data=None, **kwargs) -> Response:
    return request('put', generate_url(view, view_data), **kwargs)


def patch(view, view_data=None, **kwargs) -> Response:
    return request('patch', generate_url(view, view_data), **kwargs)


def delete(view, view_data=None, **kwargs) -> Response:
    return request('delete', generate_url(view, view_data), **kwargs)



import os
import signal
from functools import partial

from flask import Blueprint, request, current_app, jsonify, g
from flask_jwt_extended import create_access_token, create_refresh_token, jwt_refresh_token_required, get_jwt_identity, \
    current_user

import dm
from dm import defaults
from dm.domain.entities import Server, Catalog
from dm.domain.entities.user import User
from dm.utils.helpers import get_now
from dm.web.decorators import forward_or_dispatch, validate_schema
from dm.web.json_schemas import schema_healthcheck, login_post

blueprint_name = 'root'
root_bp = Blueprint(blueprint_name, __name__)


@root_bp.route('/')
def home():
    return {'message': 'Welcome to dimensigon'}


def shutdown_server():
    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:
        func = partial(os.kill, os.getppid(), signal.SIGTERM)
    current_app.logger.info('Shutting down server')
    current_app.scheduler.shutdown()
    if func is None:
        raise RuntimeError('Not running with the Werkzeug Server')
    func()


@root_bp.route('/healthcheck', methods=['GET', 'POST'])
@forward_or_dispatch
@validate_schema(POST=schema_healthcheck)
def healthcheck():
    if request.method == 'GET':
        # catalog_ver = current_app.catalog_manager.max_data_mark
        # if catalog_ver:
        #     catalog_ver = current_app.catalog_manager.max_data_mark.strftime(current_app.catalog_manager.format)
        catalog_ver = Catalog.max_catalog()
        return {"version": dm.__version__,
                "catalog_version": catalog_ver.strftime(defaults.DATEMARK_FORMAT) if catalog_ver else None,
                "scheduler": "running" if getattr(current_app.extensions.get('scheduler'), 'running',
                                                  None) else "stopped",
                "neighbours": [str(server) for server in Server.get_neighbours()],
                "services": [],
                "server": {'id': str(g.server.id),
                           'name': g.server.name}
                }
    elif request.method == 'POST':
        data = request.json

        if data.get('action') == 'stop':
            shutdown_server()
            return jsonify(''), 202
        elif data.get('action') == 'restart':
            return jsonify({'message': 'restart is not implemented'}), 500


@root_bp.route('/ping', methods=['POST'])
@forward_or_dispatch
def ping():
    req_data = request.json
    if req_data:
        req_data.update(dest_time=get_now().strftime(defaults.DATETIME_FORMAT))
        if 'servers' not in req_data:
            req_data.update(servers={})
    else:
        req_data = dict(dest_time=get_now().strftime(defaults.DATETIME_FORMAT))
    return req_data, 200


@root_bp.route('/login', methods=['POST'])
@forward_or_dispatch
@validate_schema(login_post)
def login():
    user = User.get_by_user(user=request.json.get('username', None))
    password = request.json.get('password', None)
    try:
        if not user or not user.verify_password(password):
            return {"error": "Bad username or password"}, 401
    except TypeError:
        return {"error": "Bad username or password"}, 401

    # Use create_access_token() and create_refresh_token() to create our
    # access and refresh tokens
    ret = {
        'access_token': create_access_token(identity=str(user.id), fresh=True),
        'refresh_token': create_refresh_token(identity=str(user.id))
    }
    return jsonify(ret), 200


@root_bp.route('/refresh', methods=['POST'])
@forward_or_dispatch
@jwt_refresh_token_required
def refresh():
    current_user_id = get_jwt_identity()
    ret = {
        'username': current_user.user,
        'access_token': create_access_token(identity=current_user_id, fresh=False)
    }
    return jsonify(ret), 200

@root_bp.route('/fresh-login', methods=['POST'])
@forward_or_dispatch
@validate_schema(login_post)
def fresh_login():
    username = request.json.get('username', None)
    password = request.json.get('password', None)
    if username != 'test' or password != 'test':
        return jsonify({"msg": "Bad username or password"}), 401

    new_token = create_access_token(identity=username, fresh=True)
    ret = {'access_token': new_token}
    return jsonify(ret), 200


from unittest import TestCase

from flask_jwt_extended import create_access_token
from sqlalchemy.orm.exc import NoResultFound

from dm.domain.entities import Gate, Server
from dm.web import create_app, db


class TestGate(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.headers = {"Authorization": f"Bearer {create_access_token('00000000-0000-0000-0000-000000000001')}"}

        db.create_all()
        db.session.commit()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_to_from_json(self):
        s = Server('test')
        g = Gate(server=s, dns='dns', port=8)

        g_json = g.to_json()
        self.assertEqual(g.dns, g_json['dns'])
        self.assertEqual(g.ip, g_json['ip'])
        self.assertEqual(g.port, g_json['port'])
        self.assertIsNotNone(g_json['server_id'])

        # set id to resolve server_id from gate object
        s = Server('test', id='aaaaaaaa-1234-5678-1234-56781234aaa1')
        g = Gate(server=s, dns='dns', port=8, id='aaaaaaaa-1234-5678-1234-56781234aaa2')

        g_json = g.to_json()
        self.assertDictEqual(
            {'server_id': 'aaaaaaaa-1234-5678-1234-56781234aaa1', 'dns': 'dns', 'ip': None, 'port': 8, 'hidden': False,
             'id': 'aaaaaaaa-1234-5678-1234-56781234aaa2'}, g_json)

        with self.assertRaises(NoResultFound):
            smashed = Gate.from_json(g_json)

        s = Server('test2', id='aaaaaaaa-1234-5678-1234-56781234aaa1')
        db.session.add(s)

        g_smashed = Gate.from_json(g_json)

        self.assertIsNot(g, g_smashed)
        self.assertEqual(g.dns, g_smashed.dns)
        self.assertEqual(g.ip, g_smashed.ip)
        self.assertEqual(g.port, g_smashed.port)
        self.assertIs(s, g_smashed.server)

        db.session.add(g_smashed)
        db.session.commit()

        g_json = g.to_json()

        self.assertEqual(
            {'id': 'aaaaaaaa-1234-5678-1234-56781234aaa2', 'server_id': 'aaaaaaaa-1234-5678-1234-56781234aaa1',
             'dns': 'dns', 'ip': None, 'port': 8, 'hidden': False}, g_json)

        # check from_json when persisted in database
        smashed = Gate.from_json(g_json)

        self.assertIs(g_smashed, smashed)
        self.assertEqual(g_smashed.server, smashed.server)
        self.assertEqual(g_smashed.dns, smashed.dns)
        self.assertEqual(g_smashed.ip, smashed.ip)
        self.assertEqual(g_smashed.port, smashed.port)



from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.functions import FunctionElement

from dm.utils.typos import UtcDateTime


class utcnow(FunctionElement):
    """UTCNOW() expression for multiple dialects."""
    type = UtcDateTime()


@compiles(utcnow)
def default_sql_utcnow(element, compiler, **kw):
    """Assume, by default, time zones work correctly.
    Note:
        This is a valid assumption for PostgreSQL and Oracle.
    """
    return 'CURRENT_TIMESTAMP'


@compiles(utcnow, 'mysql')
def mysql_sql_utcnow(element, compiler, **kw):
    """MySQL returns now as localtime, so we convert to UTC.
    Warning:
        MySQL does not support the use of functions for sqlalchemy
        `server_default=` values. The utcnow function must be used as
        `default=` when interacting with a MySQL server.
    """
    return "CONVERT_TZ(CURRENT_TIMESTAMP, @@session.time_zone, '+00:00')"


@compiles(utcnow, 'sqlite')
def sqlite_sql_utcnow(element, compiler, **kw):
    return "(DATETIME('NOW'))"


@compiles(utcnow, 'mssql')
def mssql_sql_utcnow(element, compiler, **kw):
    """MS SQL provides a function for the UTC datetime."""
    return 'GETUTCDATE()'


__author__ = "Joan Prat"
__copyright__ = "Copyright 2019, The Dimensigon project"
__credits__ = ["Joan Prat", "Daniel Moya"]
__license__ = ""
__version__ = "0.1-b12"
__maintainer__ = "Joan Prat"
__email__ = "joan.prat@dimensigon.com"
__status__ = "development"



import argparse
import copy
import json
import os
import shlex
from pprint import pprint

from prompt_toolkit import prompt
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.history import InMemoryHistory

import dm.dshell.network as ntwrk
from dm.dshell.argparse_raise import ArgumentParserRaise
from dm.dshell.helpers import get_history, exit_dshell
from dm.dshell.prompts.step import subprompt as step_subprompt
from dm.dshell.prompts.utils import prompt_parameter
from dm.dshell.validators import BoolValidator

form = {
    "name": dict(history=InMemoryHistory()),
    "description": dict(multiline=True, history=InMemoryHistory()),
    "stop_on_error": dict(validator=BoolValidator, history=InMemoryHistory()),
    "stop_undo_on_error": dict(validator=BoolValidator, history=InMemoryHistory()),
    "undo_on_error": dict(validator=BoolValidator, history=InMemoryHistory()),
}

parser = ArgumentParserRaise(allow_abbrev=False, prog='')
subparser = parser.add_subparsers(dest='cmd')
preview_action = subparser.add_parser('preview')
preview_action.set_defaults(func=lambda x: pprint(x))
set_action = subparser.add_parser('set')
set_action.add_argument('parameter', help="Available: " + ', '.join(form.keys()))
submit_action = subparser.add_parser('submit')
dump_action = subparser.add_parser('dump')
dump_action.add_argument('file')
exit_action = subparser.add_parser('exit')
step_parser = subparser.add_parser('step', help='adds or modifies a step')
step_parser.add_argument('id_or_type', help='modifies a step. To create a new step, specify "do" or "undo" step')

history = None

entity_name = os.path.basename(__file__).rstrip('.py')


def submit(entity):
    resp = ntwrk.post('api_1_0.orchestrationfull', json=entity)
    if not resp.ok:
        pprint(resp.msg or resp.exception)
        return False
    pprint(resp.msg)
    return True


def subprompt(entity, changed=False, ask_all=False, parent_prompt=None):
    global history

    if history is None:
        history = get_history(entity_name, InMemoryHistory())

    if ask_all:
        for k, v in form.items():
            if k not in entity:
                try:
                    if prompt_parameter(k, entity, form, f"{parent_prompt}{entity_name}('{entity['name']}')"):
                        changed = True
                except KeyboardInterrupt:
                    entity[k] = None
                except EOFError:
                    exit_dshell()

    while True:
        try:
            text = prompt(f"{parent_prompt}{entity_name}('{entity['name']}')> ", history=history,
                          auto_suggest=AutoSuggestFromHistory())
        except KeyboardInterrupt:
            continue
        except EOFError:
            exit_dshell(rc=1)
        try:
            namespace = parser.parse_args(shlex.split(text))
        except (ValueError, argparse.ArgumentError) as e:
            print(e)
            continue
        except SystemExit:
            continue

        if namespace.cmd == 'preview':
            orch = copy.deepcopy(entity)
            for s in orch['steps']:
                s.pop('orchestration_id', None)
            pprint(orch)
        elif namespace.cmd == 'set':
            if namespace.parameter not in form.keys():
                print("Not a valid parameter. Available: " + ', '.join(form.keys()))
            else:
                try:
                    if prompt_parameter(namespace.parameter, entity, form,
                                        f"{parent_prompt}{entity_name}('{entity['name']}')"):
                        changed = True
                except KeyboardInterrupt:
                    continue  # Control-C pressed. Try again.
                except EOFError:
                    exit_dshell(rc=1)
        elif namespace.cmd == 'submit':
            created = submit(entity)
            if created:
                return
        elif namespace.cmd == 'dump':
            orch = copy.deepcopy(entity)
            for s in orch['steps']:
                s.pop('orchestration_id', None)
            with open(namespace.file, 'w') as dumpfile:
                json.dump(orch, dumpfile, indent=4)
            changed = False
        elif namespace.cmd == 'step':
            if 'steps' not in entity:
                entity['steps'] = []

            step = {}
            if namespace.id_or_type not in ('do', 'undo'):
                for s in entity.get('steps', []):
                    if str(s.get('id')) == namespace.id_or_type:
                        step = s
                        break
                else:
                    print('id step does not exists in this orchestration')
                    continue

            if not step:
                # generate id
                step['id'] = str(len(entity.get('steps', [])) + 1)
                step['undo'] = namespace.id_or_type == 'undo'
                step['parent_step_ids'] = []
            changed_step = step_subprompt(step, parent_prompt=f"{parent_prompt}{entity_name}('{entity['name']}')")
            if changed_step:
                changed = True
                steps = entity.get('steps', [])
                id2step = {s['id']: s for s in steps}
                if changed_step.get('id') in id2step:
                    id2step[changed_step.get('id')].update(**changed_step)
                else:
                    entity['steps'].append(changed_step)

        elif namespace.cmd == 'exit':
            if changed:
                text = ''
                while text.lower().strip() not in ('y', 'n'):
                    text = prompt('If you exit you will lose the changes. Do you want to continue? (y/n): ')
                if text.lower().strip() == 'y':
                    return
            else:
                return



import logging
import os
import os.path
import tempfile
from subprocess import CompletedProcess
from unittest import TestCase
from unittest.mock import patch, mock_open

import dm.defaults as d
from elevator import _upgrade, HOME, DM_ROOT, FAILED_VERSIONS


class TestElevator(TestCase):

    def setUp(self) -> None:
        self.temp_dir = tempfile.gettempdir()
        self.deployable = 'dimensigon.tar.gz'
        self.version = '1.0'
        self.old_home = os.path.join(DM_ROOT, 'dimensigon')
        self.new_home = os.path.join(DM_ROOT, 'dimensigon_' + self.version)

    @patch('elevator.run')
    @patch('elevator.os.rmdir')
    @patch('elevator.os.chdir')
    @patch('elevator.daemon_running')
    @patch('elevator.start_daemon')
    @patch('elevator.stop_daemon')
    @patch('elevator.shutil')
    @patch('elevator.os.path.exists')
    @patch('elevator.get_hc')
    def test_upgrade(self, mock_get_hc, mock_exists, mock_shutil, mock_stop, mock_start, mock_daemon, mock_chdir,
                     mock_rmdir, mock_run):
        mock_daemon.return_value = True
        mock_stop.return_value = True
        mock_start.return_value = True
        mock_get_hc.side_effect = [{"version": '0.0',
                                    "elevator_version": '0.0',
                                    "catalog_version": "20190401000000000000",
                                    "scheduler": "running",
                                    "neighbours": [],
                                    "services": []
                                    },
                                   {"version": self.version,
                                    "elevator_version": '0.9',
                                    "catalog_version": "20190401000000000000",
                                    "scheduler": "running",
                                    "neighbours": [],
                                    "services": []
                                    }
                                   ]

        mock_exists.return_value = False
        mock_run.return_value = CompletedProcess((), 0, 'stdout', '')

        import elevator
        elevator.host = '127.0.0.1:5000'
        # runner = CliRunner()
        # result = runner.invoke(upgrade, ['deployable.tar.gz', '1.0'], catch_exceptions=False)
        _upgrade(dict(deployable=self.deployable, version=self.version, dm_url=f"https://127.0.0.1:{d.LOOPBACK_PORT}"))

        # extract new version
        self.assertEqual(((self.new_home,),), mock_shutil.rmtree.call_args_list[0])
        mock_shutil.unpack_archive.assert_called_once_with(self.deployable, self.temp_dir)
        self.assertEqual(((os.path.join(self.temp_dir, 'dimensigon'), self.new_home),),
                         mock_shutil.copytree.call_args_list[0])

        # copy config files and DB from old version to new version
        self.assertGreaterEqual(mock_shutil.copy2.call_count, 1)
        self.assertGreaterEqual(mock_shutil.copytree.call_count, 1)

        # migration
        mock_run.assert_called_once()

        # stop old version
        mock_stop.assert_called_once()

        # change working dir to new home
        mock_chdir.assert_called_once_with(self.new_home)

        #####################
        # start NEW version #
        #####################
        mock_start.assert_called_once_with(self.new_home, elevator.logger.level != logging.DEBUG)
        mock_daemon.assert_called_once()

    @patch('elevator.run')
    @patch('elevator.os.rmdir')
    @patch('elevator.os.chdir')
    @patch('elevator.daemon_running')
    @patch('elevator.start_daemon')
    @patch('elevator.stop_daemon')
    @patch('elevator.shutil')
    @patch('elevator.os.path.exists')
    @patch('elevator.get_hc')
    def test_upgrade_error_new_version(self, mock_get_hc, mock_exists, mock_shutil, mock_stop, mock_start, mock_daemon,
                                       mock_chdir, mock_rmdir, mock_run):
        mock_daemon.side_effect = [False, True]
        mock_stop.return_value = True
        mock_start.return_value = True
        mock_get_hc.side_effect = [{"version": '0.0',
                                    "elevator_version": '0.0',
                                    "catalog_version": "20190401000000000000",
                                    "scheduler": "running",
                                    "neighbours": [],
                                    "services": []
                                    },
                                   {"version": '0.0',
                                    "elevator_version": '0.0',
                                    "catalog_version": "20190401000000000000",
                                    "scheduler": "running",
                                    "neighbours": [],
                                    "services": []
                                    }
                                   ]

        mock_exists.return_value = False
        mock_run.return_value = CompletedProcess((), 0, 'stdout', '')

        import elevator
        elevator.host = '127.0.0.1:5000'

        # runner = CliRunner()
        # result = runner.invoke(upgrade, ['deployable.tar.gz', '1.0'], catch_exceptions=False)
        m = mock_open()
        with patch('elevator.open', m):
            _upgrade(
                dict(deployable=self.deployable, version=self.version))

        # extract new version
        self.assertEqual(((self.new_home,),), mock_shutil.rmtree.call_args_list[0])
        mock_shutil.unpack_archive.assert_called_once_with(self.deployable, self.temp_dir)
        self.assertEqual(((os.path.join(self.temp_dir, 'dimensigon'), self.new_home),),
                         mock_shutil.copytree.call_args_list[0])

        # copy config files and DB from old version to new version
        self.assertGreaterEqual(mock_shutil.copy2.call_count, 1)
        self.assertGreaterEqual(mock_shutil.copytree.call_count, 1)

        # migration
        mock_run.assert_called_once()

        # stop old version
        # mock_stop.assert_called_once()

        # change working dir to new home
        # mock_chdir.assert_called_once_with(self.new_home)

        #####################
        # start NEW version #
        #####################
        # mock_start.assert_called_once_with(self.new_home, logging.root.level != logging.DEBUG)
        # mock_daemon.assert_called_once()

        # kill daemon if running
        self.assertEqual(2, mock_stop.call_count)

        self.assertListEqual([((self.new_home,),), ((HOME,),)], mock_chdir.call_args_list)

        m.assert_called_once_with(FAILED_VERSIONS, 'a')
        handle = m()
        handle.write.assert_called_once_with(self.version + '\n')

        # start old version
        self.assertListEqual(
            [((self.new_home, elevator.logger.level != logging.DEBUG),),
             ((self.old_home, elevator.logger.level != logging.DEBUG),)],
            mock_start.call_args_list)

        self.assertEqual(2, mock_daemon.call_count)



import base64
import functools
import ipaddress
import socket
import typing as t

import requests
import rsa
from flask import current_app, url_for, g
from jsonschema import validate

from dm import defaults
from dm.domain.entities import Server, Scope, User
from dm.network.exceptions import NotValidMessage
from dm.use_cases.helpers import get_servers_from_scope
from dm.use_cases.lock import lock_scope
from dm.utils.helpers import get_now
from dm.web import db, errors
from dm.web.network import unpack_msg, pack_msg, unpack_msg2, pack_msg2

if t.TYPE_CHECKING:
    import flask


def save_if_hidden_ip(remote_addr: str, server: Server):
    ip = ipaddress.ip_address(remote_addr)
    if not ip.is_loopback:
        def get_ip(dns: str):
            if dns is None:
                return
            else:
                try:
                    return ipaddress.ip_address(socket.gethostbyname(dns))
                except:
                    return

        gate = [gate for gate in server.gates if ip in (gate.ip, get_ip(gate.dns))]
        if not gate:
            hidden_gates = server.hidden_gates
            if hidden_gates:
                for hg in hidden_gates:
                    hg.ip = remote_addr
            else:
                for port in set([gate.port for gate in server.gates]):
                    gate = server.add_new_gate(dns_or_ip=remote_addr, port=port, hidden=True)
                    db.session.add(gate)
            db.session.commit()


def _proxy_request(request: 'flask.Request', destination: Server, verify=False) -> requests.Response:
    url = destination.url() + request.full_path
    req_data = request.get_json()

    if request.path == '/ping':
        server_data = {'id': str(g.server.id), 'name': g.server.name,
                       'time': get_now().strftime(defaults.DATETIME_FORMAT)}
        if req_data:
            if 'servers' not in req_data:
                req_data['servers'] = {}
            req_data['servers'].update({len(req_data['servers']) + 1: server_data})
        else:
            req_data = dict(servers={1: server_data})
    kwargs = {
        'json': req_data,
        'allow_redirects': False
    }

    headers = {key.lower(): value for key, value in request.headers.items()}

    # Let requests reset the host for us.
    if 'host' in headers:
        del headers['host']

    headers['d-source'] = headers.get('d-source', '') + ':' + str(g.server.id)

    kwargs['headers'] = headers

    cookies = request.cookies

    kwargs['cookies'] = cookies

    return requests.request(request.method, url, verify=verify, **kwargs)


def forward_or_dispatch(func):
    from flask import request

    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        destination_id = None
        if 'D-Destination' in request.headers:
            destination_id = request.headers['D-Destination']
        else:
            # Get information from content
            # Code Compatibility. Use D-Destination header instead
            data = request.get_json()
            if data is not None and 'destination' in data:
                destination_id = data.get('destination')

        if destination_id and destination_id != str(g.server.id):
            destination = Server.query.get(destination_id)
            if destination is None:
                return errors.format_error_response(errors.EntityNotFound('Server', destination_id))
            try:
                resp = _proxy_request(request=request, destination=destination)
            except requests.exceptions.RequestException as e:
                return errors.format_error_response(errors.ProxyForwardingError(destination, e))
            else:
                return resp.content, resp.status_code, dict(resp.headers)

        else:
            # not call get_or_404 to allow make requests without D-Source header
            source_id, proxies = (request.headers.get('D-Source', '') + ':').split(':', 1)
            proxies = proxies.strip(':')
            source = db.session.query(Server).get(source_id)
            # check hidden ip on server
            if proxies:
                lp = proxies.split(':')
                neighbour = db.session.query(Server).get(lp[-1])
                if neighbour:
                    save_if_hidden_ip(request.remote_addr, neighbour)
            if not source:
                source = request.headers.get('D-Source') or request.remote_addr
            elif not proxies:
                save_if_hidden_ip(request.remote_addr, source)
            g.source = source

            value = func(*args, **kwargs)
            return value

    return wrapper_decorator


def securizer(func):
    from flask import request
    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):

        # cipher_key = session.get('cipher_key', None)
        cipher_key = None
        securizer_method = None

        if 'D-Securizer' in request.headers:
            securizer_method = request.headers.get('D-Securizer')
            if securizer_method == 'plain' and not current_app.config.get('SECURIZER_PLAIN', False):
                    return {'error': 'plain data is not allowed'}, 406

        if request.method != 'GET':
            if request.is_json:
                if securizer_method == 'plain':
                    pass
                else:
                    g.original_json = request.get_json()
                    cipher_key = base64.b64decode(request.json.get('key')) if 'key' in request.json else cipher_key
                    # session['cipher_key'] = cipher_key
                    try:
                        if request.path == url_for('api_1_0.join'):
                            temp_pub_key = rsa.PublicKey.load_pkcs1(request.json.pop('my_pub_key').encode('ascii'))
                            data = unpack_msg2(data=request.get_json(),
                                               pub_key=temp_pub_key,
                                               priv_key=getattr(getattr(g, 'dimension', None), 'private', None),
                                               cipher_key=cipher_key)
                        else:
                            data = unpack_msg(data=request.get_json())
                    except (rsa.pkcs1.VerificationError, NotValidMessage) as e:
                        return {'error': str(e),
                                'message': request.get_json()}, 400
                    if current_app.config.get('SECURIZER', None) and data:
                        # fill json request with unpacked data
                        for key in list(request.json.keys()):
                            request.json.pop(key)
                        for key, val in data.items():
                            request.json[key] = val

            else:
                if request.data:
                    return {'error': 'Content Type must be application/json'}, 400

        rest = tuple()
        rv = func(*args, **kwargs)
        if isinstance(rv, tuple):
            len_rv = len(rv)

            if len_rv >= 2:
                rest = rv[1:]
                rv = rv[0]

        if rv is None:
            return rv

        if isinstance(rv, dict):

            if request.path == url_for('api_1_0.join'):
                rv = pack_msg2(data=rv, pub_key=temp_pub_key,
                               priv_key=getattr(getattr(g, 'dimension', None), 'private', None),
                               cipher_key=cipher_key)
            else:
                if securizer_method == 'plain' and current_app.config.get('SECURIZER_PLAIN', False):
                    pass
                else:
                    rv = pack_msg(data=rv)

        if isinstance(rv, list):
            if securizer_method == 'plain' and current_app.config.get('SECURIZER_PLAIN', False):
                pass
            else:
                rv = pack_msg(data=rv)

        if rest:
            rv = (rv,) + rest
        return rv

    return wrapper_decorator


def validate_schema(schema_name=None, **methods):
    from flask import request
    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kw):
            schema = methods.get(request.method.upper()) or methods.get(request.method.lower()) or schema_name
            if schema:
                validate(request.json, schema)
            return f(*args, **kw)

        return wrapper

    return decorator


def lock_catalog(f):
    @functools.wraps(f)
    def wrapper(*args, **kw):
        servers = get_servers_from_scope(Scope.CATALOG)
        with lock_scope(Scope.CATALOG, servers):
            try:
                ret = f(*args, **kw)
            except Exception as e:
                db.session.rollback()
                raise
        return ret

    return wrapper


def rollback_on_error(f):
    @functools.wraps(f)
    def wrapper(*args, **kw):
        try:
            return f(*args, **kw)
        except Exception:
            db.session.rollback()
            raise

    return wrapper


def run_as(username: str):
    from flask import Flask
    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            # app resolution
            app = None
            if 'app' in kwargs:
                app = kwargs['app']
            else:
                for a in args:
                    if isinstance(a, Flask):
                        app = a
            if app:
                ctx = app.app_context()
                ctx.push()
            try:
                user = User.get_by_user(username)
                if user is None:
                    raise errors.EntityNotFound("User", username, ['name'])
                from flask_jwt_extended.utils import ctx_stack
                ctx_stack.top.jwt_user = user
                jwt = {'identity': str(user.id)}
                ctx_stack.top.jwt = jwt
                return f(*args, **kwargs)
            finally:
                if app:
                    ctx.pop()

        return wrapper

    return decorator



from unittest import TestCase

from dateutil.tz import tzlocal
from flask import url_for
from flask_jwt_extended import create_access_token

from dm import defaults
from dm.domain.entities import ActionTemplate, bypass_datamark_update, Server, User
from dm.network.auth import HTTPBearerAuth
from dm.web import create_app, db


class TestActionTemplate(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create a temporary file to isolate the database for each test
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client(use_cookies=True)
        self.auth = HTTPBearerAuth(create_access_token('00000000-0000-0000-0000-000000000001'))
        db.create_all()
        Server.set_initial()
        User.set_initial()
        self.at1_json = {"id": "aaaaaaaa-1234-5678-1234-56781234aaa1",
                         "action_type": "SHELL",
                         "code": "mkdir {dir}",
                         "last_modified_at": defaults.INITIAL_DATEMARK.strftime(defaults.DATEMARK_FORMAT),
                         "name": "mkdir",
                         "version": 1
                         }
        self.at2_json = {"id": "aaaaaaaa-1234-5678-1234-56781234aaa2",
                         "action_type": "SHELL",
                         "code": "rmdir {dir}",
                         "last_modified_at": defaults.INITIAL_DATEMARK.strftime(defaults.DATEMARK_FORMAT),
                         "expected_stdout": 'output',
                         "expected_stderr": 'err',
                         "expected_rc": 0,
                         "name": "rmdir",
                         "parameters": {'param1': 1},
                         "system_kwargs": {'kwarg1': 1},
                         'pre_process': 'pre_process',
                         'post_process': 'post_process',
                         "version": 1
                         }

        with bypass_datamark_update():
            at1 = ActionTemplate.from_json(self.at1_json)
            at2 = ActionTemplate.from_json(self.at2_json)

            db.session.add_all([at1, at2])
            db.session.commit()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_action_template_list(self):
        response = self.client.get(url_for('api_1_0.actiontemplatelist'), headers=self.auth.header)

        self.assertListEqual([
            {"id": "aaaaaaaa-1234-5678-1234-56781234aaa1",
             "action_type": "SHELL",
             "code": "mkdir {dir}",
             "last_modified_at": defaults.INITIAL_DATEMARK.astimezone(tzlocal()).strftime(defaults.DATEMARK_FORMAT),
             "name": "mkdir",
             "version": 1
             },
            {"id": "aaaaaaaa-1234-5678-1234-56781234aaa2",
             "action_type": "SHELL",
             "code": "rmdir {dir}",
             "last_modified_at": defaults.INITIAL_DATEMARK.astimezone(tzlocal()).strftime(defaults.DATEMARK_FORMAT),
             "expected_stdout": 'output',
             "expected_stderr": 'err',
             "expected_rc": 0,
             "name": "rmdir",
             "parameters": {'param1': 1},
             "system_kwargs": {'kwarg1': 1},
             'pre_process': 'pre_process',
             'post_process': 'post_process',
             "version": 1
             }
        ], response.get_json())

    def test_action_template(self):
        response = self.client.get(
            url_for('api_1_0.actiontemplateresource', action_template_id="aaaaaaaa-1234-5678-1234-56781234aaa1"),
            headers=self.auth.header)

        self.assertDictEqual(
            {"id": "aaaaaaaa-1234-5678-1234-56781234aaa1",
             "action_type": "SHELL",
             "code": "mkdir {dir}",
             "last_modified_at": defaults.INITIAL_DATEMARK.astimezone(tzlocal()).strftime(defaults.DATEMARK_FORMAT),
             "name": "mkdir",
             "version": 1
             }, response.get_json())




from unittest import TestCase

from dm.utils.var_context import VarContext


class TestVarContext(TestCase):

    def test_get(self):
        defaults = dict(d=4, o=4, s=4, v=4)
        variables = dict(v=1)
        vc = VarContext(defaults=defaults, variables=variables)

        self.assertEqual(4, vc.get('d'))

        self.assertIsNone(vc.get('x'))
        self.assertEqual(2, vc.get('x', 2))

    def test_create_new_ctx(self):
        initials = dict(i=3, d=3, v=3)
        variables = dict(v=1)
        vc = VarContext(initials=initials, variables=variables)

        new_vs = vc.create_new_ctx(defaults=dict(d=2, v=2))

        self.assertEqual(1, new_vs.get('v'))
        self.assertEqual(2, new_vs.get('d'))
        self.assertEqual(3, new_vs.get('i'))

        vc.set('v', 4)
        self.assertEqual(4, new_vs.get('v'))

    def test_recursive(self):
        initials = dict(a='{{b}}')
        defaults = dict(b='{{c}}')
        variables = dict(c=1)
        vc = VarContext(initials=initials, defaults=defaults, variables=variables)

        self.assertEqual(1, vc.get('a'))

        vc.set('c', '{{a}}')
        with self.assertRaises(RecursionError):
            vc.get('a')

    def test_in(self):
        initials = dict(i=3)
        defaults = dict(d=2)
        variables = dict(v=1)
        vc = VarContext(initials=initials, defaults=defaults, variables=variables)

        self.assertIn('i', vc)
        self.assertIn('d', vc)
        self.assertIn('v', vc)

    def test_dict(self):
        initials = dict(i=3, d=3, v=3)
        defaults = dict(d=2, v=2)
        variables = dict(v=1)
        vc = VarContext(initials=initials, defaults=defaults, variables=variables)

        self.assertDictEqual(dict(i=3, d=2, v=1), dict(vc))
        self.assertEqual(3, len(vc))

    def test_empty(self):
        vc = VarContext()
        self.assertEqual(0, len(vc))
        self.assertEqual({}, dict(vc))
        self.assertIsNone(vc.get('x'))

    def test_globals(self):
        globals = dict(g=4, i=4, d=4, v=4)
        initials = dict(i=3, d=3, v=3)
        defaults = dict(d=2, v=2)
        variables = dict(v=1)
        vc = VarContext(globals=globals, initials=initials, defaults=defaults, variables=variables)

        self.assertDictEqual(dict(i=3, d=2, v=1), dict(vc))
        self.assertEqual(3, len(vc))

        self.assertDictEqual(dict(g=4, i=4, d=4, v=4), dict(vc.globals))
        #
        # with self.assertRaises(TypeError):
        #     vc.globals['g'] = 0



import datetime as dt
import functools
import re
from functools import partial
from unittest import TestCase
from unittest.mock import patch

import responses
from aioresponses import aioresponses, CallbackResult

from dimensigon import join, Gate, token as generate_token, Locker, User
from dm.domain.entities import Server, Dimension, Catalog
from dm.utils.helpers import generate_dimension
from dm.web import create_app, db


class TestApi(TestCase):

    @patch('dm.domain.entities.get_now')
    def setUp(self, mock_now):
        """Create and configure a new app instance for each test."""
        mock_now.return_value = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)
        # create the app with common test config
        self.app_join = create_app('test')
        self.app_join.config['SECURIZER'] = True
        self.app_join.config['SERVER_NAME'] = 'new'
        with self.app_join.app_context():
            db.create_all()
            Locker.set_initial()
            User.set_initial()
            Server.set_initial()
            me = Server.get_current()
            # remove Gates
            [db.session.delete(g) for g in me.gates]
            g = Gate(me, port=8000, dns='new')
            db.session.add(g)
            db.session.commit()
            self.server_new = Server.get_current().to_json()

        self.app = create_app('test')
        self.app.config['SERVER_NAME'] = 'node1'
        self.app.config['SECURIZER'] = True
        mock_now.return_value = dt.datetime(2019, 3, 1, tzinfo=dt.timezone.utc)
        with self.app.app_context():
            db.create_all()
            Locker.set_initial()
            User.set_initial()
            Server.set_initial()
            me = Server.get_current()
            # remove Gates
            [db.session.delete(g) for g in me.gates]
            g = Gate(me, port=8000, dns='node1')
            db.session.add(g)
            db.session.commit()
            dim = generate_dimension('test')
            dim.current = True
            db.session.add(dim)
            db.session.commit()
            self.dimension = dim.to_json()
            self.server_node1 = Server.get_current().to_json()

        self.client = self.app.test_client()

    def tearDown(self) -> None:
        with self.app_join.app_context():
            db.session.remove()
            db.drop_all()
        with self.app.app_context():
            db.session.remove()
            db.drop_all()

    @patch('dm.domain.entities.route.check_host')
    @patch('dm.domain.entities.get_now')
    # @patch('dimensigon.rsa.newkeys')
    @aioresponses()
    def test_join_command(self, mock_now, mock_check, m):
        mock_now.return_value = dt.datetime(2019, 5, 1, tzinfo=dt.timezone.utc)

        mock_check.return_value = True

        def callback_client(method, client, url, **kwargs):
            kwargs.pop('allow_redirects')

            func = getattr(client, method.lower())
            r = func(url.path, headers=kwargs['headers'], json=kwargs['json'])

            return CallbackResult(method.upper(), status=r.status_code, body=r.data, content_type=r.content_type,
                                  headers=r.headers)

        m.post(re.compile('^https?://node1.*'),
               callback=functools.partial(callback_client, 'POST', self.client), repeat=True)

        def request_callback(request, client):
            method_func = getattr(client, request.method.lower())
            resp = method_func(request.path_url, data=request.body, headers=dict(request.headers))

            return resp.status_code, resp.headers, resp.data

        with responses.RequestsMock() as rsps:
            rsps.add_callback(responses.POST,
                              re.compile('^https?://node1.*'),
                              callback=partial(request_callback, client=self.client))
            rsps.add_callback(responses.GET,
                              re.compile('^https?://node1.*'),
                              callback=partial(request_callback, client=self.client))

            with self.app.app_context():
                runner = self.app.test_cli_runner()
                result = runner.invoke(generate_token, [], catch_exceptions=False)
                token = result.stdout.strip()

            runner = self.app_join.test_cli_runner()

            result = runner.invoke(join, ['--no-ssl', 'node1:5000', token], catch_exceptions=False)

        with self.app.app_context():
            # check if new server created
            server_new = Server.query.get(self.server_new.get('id'))
            self.assertEqual(server_new.last_modified_at, dt.datetime(2019, 5, 1, tzinfo=dt.timezone.utc))

        with self.app_join.app_context():
            self.assertDictEqual(self.dimension, Dimension.get_current().to_json())
            s = Server.query.get(self.server_node1.get('id'))
            self.assertEqual(0, s.route.cost)
            self.assertListEqual([('Gate',), ('Server',), ('User',)],
                                 db.session.query(Catalog.entity).order_by('entity').all())
            self.assertEqual(dt.datetime(2019, 5, 1, tzinfo=dt.timezone.utc),
                             Catalog.query.filter_by(entity='Server').one().last_modified_at)



from unittest import TestCase

from flask import url_for
from flask_jwt_extended import create_access_token
from werkzeug.exceptions import InternalServerError

from dm.domain.entities import User
from dm.domain.entities.bootstrap import set_initial
from dm.network.auth import HTTPBearerAuth
from dm.web import create_app, db, errors


class TestApi(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app.config['DEBUG'] = True
        self.app.config['TESTING'] = False
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()

        db.create_all()
        set_initial()
        u = User('test')
        db.session.add(u)
        self.auth = HTTPBearerAuth(create_access_token(u.id))

        db.session.commit()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_validation_error(self):
        def raise_error():
            raise errors.ValidationError('error content', schema={})

        self.app.add_url_rule('/error', 'error', raise_error)
        resp = self.client.get(url_for('error'), json={}, headers=self.auth.header)

        self.assertEqual(400, resp.status_code)
        self.assertDictEqual({'error': {'type': 'ValidationError', 'message': 'error content', 'schema': {}}},
                             resp.get_json())

    def test_base_error(self):
        def raise_error():
            raise errors.GenericError('error content', some='payload')

        self.app.add_url_rule('/error', 'error', raise_error)

        resp = self.client.get(url_for('error'), json={}, headers=self.auth.header)

        self.assertEqual(400, resp.status_code)
        self.assertDictEqual({'error': {'type': 'GenericError', 'message': "error content", 'some': 'payload'}},
                             resp.get_json())

    def test_internal_server_error(self):
        self.app.config['PROPAGATE_EXCEPTIONS'] = False

        def raise_error():
            raise RuntimeError('error content')

        self.app.add_url_rule('/error', 'error', raise_error)
        resp = self.client.get(url_for('error'), json={}, headers=self.auth.header)

        self.assertEqual(500, resp.status_code)
        self.assertDictEqual({'error': {'type': "InternalServerError", "message": InternalServerError.description}},
                             resp.get_json())



from unittest import TestCase
from unittest.mock import patch

from dm import defaults
from dm.domain.entities import Server
from dm.web import create_app, db


class TestServer(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        db.create_all()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_create_server(self):
        s = Server('test')
        self.assertEqual(0, len(s.gates))

        s = Server('test', dns_or_ip='dns')
        self.assertEqual(1, len(s.gates))
        self.assertIsNone(s.gates[0].ip)
        self.assertEqual('dns', s.gates[0].dns)
        self.assertEqual(defaults.DEFAULT_PORT, s.gates[0].port)

        s = Server('test', port=5000)
        self.assertEqual(1, len(s.gates))
        self.assertIsNone(s.gates[0].ip)
        self.assertEqual('test', s.gates[0].dns)
        self.assertEqual(5000, s.gates[0].port)

        s = Server('test', dns_or_ip='dns', port=5000)
        self.assertEqual(1, len(s.gates))
        self.assertIsNone(s.gates[0].ip)
        self.assertEqual('dns', s.gates[0].dns)
        self.assertEqual(5000, s.gates[0].port)

        # create with gates
        s = Server('test', gates=[('gdns', 6000)])
        self.assertEqual(1, len(s.gates))
        self.assertIsNone(s.gates[0].ip)
        self.assertEqual('gdns', s.gates[0].dns)
        self.assertEqual(6000, s.gates[0].port)

        s = Server('test', dns_or_ip='dns', gates=[('gdns', 6000)])
        self.assertEqual(2, len(s.gates))
        self.assertIsNone(s.gates[0].ip)
        self.assertEqual('dns', s.gates[0].dns)
        self.assertEqual(defaults.DEFAULT_PORT, s.gates[0].port)
        self.assertIsNone(s.gates[1].ip)
        self.assertEqual('gdns', s.gates[1].dns)
        self.assertEqual(6000, s.gates[1].port)

        s = Server('test', port=5000, gates=[('gdns', 6000)])
        self.assertEqual(2, len(s.gates))
        self.assertIsNone(s.gates[0].ip)
        self.assertEqual('test', s.gates[0].dns)
        self.assertEqual(5000, s.gates[0].port)
        self.assertIsNone(s.gates[1].ip)
        self.assertEqual('gdns', s.gates[1].dns)
        self.assertEqual(6000, s.gates[1].port)

        s = Server('test', dns_or_ip='dns', port=5000, gates=[('gdns', 6000)])
        self.assertEqual(2, len(s.gates))
        self.assertIsNone(s.gates[0].ip)
        self.assertEqual('dns', s.gates[0].dns)
        self.assertEqual(5000, s.gates[0].port)
        self.assertIsNone(s.gates[1].ip)
        self.assertEqual('gdns', s.gates[1].dns)
        self.assertEqual(6000, s.gates[1].port)

    @patch('dm.domain.entities.server.Gate')
    def test_create_server_dict_gate(self, mock_gate):
        dest = Server('dest', gates=[{'id': 1}])

        mock_gate.from_json.called_once_with({'id': 1})

    @patch('dm.domain.entities.base.uuid.uuid4')
    def test_to_from_json(self, mock_uuid):
        mock_uuid.side_effect = ['22cd859d-ee91-4079-a112-000000000001',
                                 '22cd859d-ee91-4079-a112-000000000002',
                                 '22cd859d-ee91-4079-a112-000000000003']

        s = Server('server', dns_or_ip='dns', gates=[('gdns', 6000)])
        self.assertDictEqual({'id': '22cd859d-ee91-4079-a112-000000000001', 'name': 'server', 'granules': []},
                             s.to_json())
        db.session.add(s)
        db.session.commit()

        smashed = Server.from_json(s.to_json())

        self.assertIs(s, smashed)
        self.assertEqual(s.id, smashed.id)
        self.assertEqual(s.name, smashed.name)
        self.assertEqual(s.granules, smashed.granules)
        self.assertEqual(s.last_modified_at, smashed.last_modified_at)



import copy
import re
import typing as t
from enum import Enum

from dm.domain.entities.base import UUIDistributedEntityMixin, SoftDeleteMixin
from dm.utils import typos
from dm.web import db
from dm.web.helpers import QueryWithSoftDelete

if t.TYPE_CHECKING:
    from dm.domain.entities import Server


class Mode(Enum):
    REPO_MIRROR = 0
    REPO_ROOT = 1
    MIRROR = 2
    FOLDER = 3


class Log(db.Model, UUIDistributedEntityMixin, SoftDeleteMixin):
    __tablename__ = 'D_log'

    src_server_id = db.Column(typos.UUID, db.ForeignKey('D_server.id'), nullable=False)
    target = db.Column(db.Text, nullable=False)
    include = db.Column(db.Text)
    exclude = db.Column(db.Text)
    dst_server_id = db.Column(typos.UUID, db.ForeignKey('D_server.id'), nullable=False)
    mode = db.Column(typos.Enum(Mode))
    dest_folder = db.Column(db.Text)
    recursive = db.Column(db.Boolean, default=False)
    _old_target = db.Column("$$target", db.Text)

    source_server = db.relationship("Server", foreign_keys=[src_server_id])
    destination_server = db.relationship("Server", foreign_keys=[dst_server_id])

    __table_args__ = (db.UniqueConstraint('src_server_id', 'target', 'dst_server_id'),)

    query_class = QueryWithSoftDelete

    def __init__(self, source_server: 'Server', target: str, destination_server: 'Server', dest_folder=None,
                 include=None, exclude=None, recursive=False, mode=Mode.REPO_MIRROR, **kwargs):
        UUIDistributedEntityMixin.__init__(self, **kwargs)
        SoftDeleteMixin.__init__(self, **kwargs)

        self.source_server = source_server
        self.target = target
        self.destination_server = destination_server
        self.dest_folder = dest_folder
        if self.dest_folder is None:
            self.mode = mode
        else:
            self.mode = Mode.FOLDER
        self.include = include
        self._re_include = re.compile(self.include or '')
        self.exclude = exclude
        self._re_exclude = re.compile(self.exclude or '^$')
        self.recursive = recursive

    def __str__(self):
        return f"{self.source_server}:{self.target} -> {self.destination_server}:{self.dest_folder}"

    def to_json(self, human=False, delete_data=True, include: t.List[str] = None, exclude: t.List[str] = None):
        data = super().to_json()
        if self.source_server.id is None or self.destination_server.id is None:
            raise RuntimeError('Set ids for servers before')
        data.update(target=self.target, include=self.include,
                    exclude=self.exclude, dest_folder=self.dest_folder,
                    recursive=self.recursive, mode=self.mode.name, deleted=self.deleted, _old_target=self._old_target)
        if human:
            data.update(src_server=str(self.source_server.name), dst_server=str(self.destination_server.name))
        else:
            data.update(src_server_id=str(self.source_server.id), dst_server_id=str(self.destination_server.id))

        if not delete_data:
            data.pop('deleted', None)
            data = {k: v for k, v in data.items() if not k.startswith(SoftDeleteMixin.__prefix__)}

        if include:
            data = {k: v for k, v in data.items() if k in include}
        if exclude:
            data = {k: v for k, v in data.items() if k not in exclude}
        return data

    @classmethod
    def from_json(cls, kwargs) -> 'Log':
        from dm.domain.entities import Server
        kwargs = copy.deepcopy(kwargs)
        kwargs['mode'] = Mode[kwargs.get('mode')]
        kwargs['source_server'] = db.session.query(Server).get(kwargs.pop('src_server_id'))
        kwargs['destination_server'] = db.session.query(Server).get(kwargs.pop('dst_server_id'))
        return super().from_json(kwargs)



import base64
from unittest import TestCase
from unittest.mock import patch, mock_open

from flask import url_for
from flask_jwt_extended import create_access_token

from dm.domain.entities.bootstrap import set_initial
from dm.network.auth import HTTPBearerAuth
from dm.web import create_app, db


class Test(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.auth = HTTPBearerAuth(create_access_token('00000000-0000-0000-0000-000000000001'))
        db.create_all()
        set_initial()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    @patch('dm.web.api_1_0.urls.use_cases.open', mock_open(read_data=b'bibble'))
    @patch('dm.web.api_1_0.urls.use_cases.os.listdir')
    def test_software_dimensigon(self, mock_os):
        with patch('dm.web.api_1_0.urls.use_cases.dm.__version__', '1.2.0'):
            mock_os.return_value = ['file1', 'dimensigon-v1.0.0.tar.gz', 'dimensigon-v1.2.0.tar.gz',
                                    'dimensigon-v1.1.0.tar.gz']
            resp = self.client.get(url_for('api_1_0.software_dimensigon'), headers=self.auth.header)

            self.assertEqual(200, resp.status_code)

            self.assertDictEqual({'filename': 'dimensigon-v1.2.0.tar.gz', 'version': 'v1.2.0',
                                  'content': base64.b64encode(b'bibble').decode('ascii')}, resp.get_json())

            # version on filesystem bigger than actual
            mock_os.return_value = ['file1', 'dimensigon-1.2.0.tar.gz', 'dimensigon-1.3.0.tar.gz',
                                    'dimensigon-1.0.0.tar.gz']
            resp = self.client.get(url_for('api_1_0.software_dimensigon'), headers=self.auth.header)

            self.assertEqual(200, resp.status_code)

            self.assertDictEqual({'filename': 'dimensigon-1.3.0.tar.gz', 'version': '1.3.0',
                                  'content': base64.b64encode(b'bibble').decode('ascii')}, resp.get_json())

            # version on filesystem lower than actual
            mock_os.return_value = ['file1', 'dimensigon-v1.0.0.tar.gz', 'dimensigon-v0.9.0.tar.gz']
            resp = self.client.get(url_for('api_1_0.software_dimensigon'), headers=self.auth.header)

            self.assertEqual(204, resp.status_code)



import typing as t
from contextlib import contextmanager
from uuid import UUID

import aiohttp
from flask import current_app
from flask_jwt_extended import create_access_token, get_jwt_identity
from sqlalchemy.orm import sessionmaker

from dm.domain.entities import Server, Catalog
from dm.domain.entities.locker import Scope
from dm.network.auth import HTTPBearerAuth
from dm.use_cases.helpers import get_servers_from_scope
from dm.utils.asyncio import run, create_task
from dm.utils.helpers import is_iterable_not_string
from dm.utils.typos import Id
from dm.web import errors, db
from dm.web.network import async_post, Response


async def request_locker(servers: t.Union[Server, t.List[Server]], action, scope, applicant, auth=None,
                         datemark=None) -> t.List[Response]:
    tasks = []
    server_responses = []
    if is_iterable_not_string(servers):
        it = servers
    else:
        it = [servers]
    payload = dict(scope=scope.name, applicant=applicant)
    if datemark:
        payload.update(datemark=datemark)
    async with aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(ssl=current_app.config['SSL_VERIFY'])) as session:
        for server in it:
            tasks.append(create_task(async_post(server, 'api_1_0.locker_' + action, session=session,
                                                json=payload,
                                                auth=auth)))
        for server in it:
            r = await tasks.pop(0)
            server_responses.append(r)

    return server_responses


def lock_unlock(action: str, scope: Scope, servers: t.List[Server], applicant=None):
    """

    Parameters
    ----------
    action
        'U' for unlocking and 'L' for locking
    scope
        scope of the lock
    servers
        servers to ask for a lock

    Raises
    ------
    Raises an error if something went wrong

    Returns
    -------
    None
        returns none if all went as expected.
    """

    assert action in 'UL'

    applicant = applicant or [str(s.id) for s in servers]

    token = create_access_token(get_jwt_identity())
    auth = HTTPBearerAuth(token)
    if action == 'U':
        pool_responses = run(
            request_locker(servers=servers, scope=scope, action='unlock', applicant=applicant, auth=auth))

        if len(servers) == len([r for r in pool_responses if r.code in (200, 210)]):
            return
    else:
        action = 'P'
        catalog_ver = Catalog.max_catalog(str)
        pool_responses = run(
            request_locker(servers=servers, scope=scope, action='prevent', applicant=applicant, datemark=catalog_ver, auth=auth))

        if len(servers) == len([r for r in pool_responses if r.code in (200, 210)]):
            action = 'L'
            pool_responses = run(
                request_locker(servers=servers, scope=scope, action='lock', applicant=applicant, auth=auth))
            if len(servers) == len([r for r in pool_responses if r.code in (200, 210)]):
                return

    raise errors.LockError(scope, action, [r for r in pool_responses if r.code not in (200, 210)])


def lock(scope: Scope, servers: t.List[Server] = None, applicant=None) -> UUID:
    """
    locks the Locker if allowed
    Parameters
    ----------
    scope
        scope that lock will affect.
    servers
        if scope set to Scope.ORCHESTRATION,
    applicant
        identifier of the lock
    Returns
    -------
    Result
    """

    if servers is not None:
        servers = servers if is_iterable_not_string(servers) else [servers]
        if len(servers) == 0:
            servers = None

    servers = servers or get_servers_from_scope(scope)

    if len(servers) == 0:
        raise RuntimeError('no server to lock')

    applicant = applicant if applicant is not None else [str(s.id) for s in servers]
    try:
        lock_unlock(action='L', scope=scope, servers=servers, applicant=applicant)
    # exception goes here because we need the applicant to unlock already locked servers
    except errors.LockError as e:
        error_servers = [r.server for r in e.responses]
        locked_servers = list(set(server for server in servers) - set(error_servers))
        lock_unlock('U', scope, servers=locked_servers, applicant=applicant)
        raise e
    return applicant


def unlock(scope: Scope, applicant, servers: t.Union[t.List[Server], t.List[Id]] =None):
    """
    unlocks the Locker if allowed
    Parameters
    ----------
    scope

    Returns
    -------

    """
    if scope.ORCHESTRATION == scope and servers is None:
        raise ValueError('servers must be set')

    s = None
    if servers:
        if not isinstance(servers[0], Server):
            engine = db.get_engine()
            Session = sessionmaker(bind=engine)
            s = Session()
            servers = [s.session.query(Server).get(s) for s in servers]
    servers = servers or get_servers_from_scope(scope)

    if not servers:
        if s:
            s.close()
        raise RuntimeError('no server to unlock')

    lock_unlock(action='U', scope=scope, servers=servers, applicant=applicant)
    if s:
        s.close()


@contextmanager
def lock_scope(scope: Scope, servers: t.Union[t.List[Server], Server] = None):
    if servers is not None:
        servers = servers if is_iterable_not_string(servers) else [servers]
        if len(servers) == 0:
            servers = None

    servers = servers or get_servers_from_scope(scope)

    applicant = lock(scope, servers)
    current_app.logger.debug(f"Lock on {scope.name} acquired on servers : {[s.name for s in servers]}")
    try:
        yield applicant
    finally:
        unlock(scope, servers=servers, applicant=applicant)
        current_app.logger.debug(f"Lock on {scope.name} released")



from unittest import TestCase

from dm.utils.helpers import convert, get_distributed_entities


class TestConvert(TestCase):

    def test_convert(self):
        d = {'param1': 'value1', 'param2': {'test': 'test_value', 'subparam1': {'level3': 3}}}

        o = convert(d)
        self.assertEqual(o.param2.subparam1.level3, 3)

        self.assertDictEqual(o.param2.subparam1, {'level3': 3})


    def test_get_distributed_entities(self):
        import dm.domain.entities
        entities = get_distributed_entities()

        for name, cls in entities:
            self.assertTrue(hasattr(cls, 'last_modified_at'))
            self.assertTrue(name in dm.domain.entities.__all__)



from flask import request
from flask_jwt_extended import jwt_required
from flask_restful import Resource

from dm.domain.entities import Step, Orchestration, ActionTemplate, ActionType
from dm.web import db, errors
from dm.web.decorators import securizer, forward_or_dispatch, validate_schema, lock_catalog
from dm.web.helpers import filter_query
from dm.web.json_schemas import step_post, step_put, step_patch


class StepList(Resource):

    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self):
        query = filter_query(Step, request.args)
        return [s.to_json() for s in query.all()]

    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(step_post)
    @lock_catalog
    def post(self):
        json_data = request.get_json()
        if not isinstance(json_data, list):
            json_data = [json_data]

        new_id_steps = []
        rid2step = {}
        dependencies = {}
        o: Orchestration = None
        for json_step in json_data:
            rid = json_step.pop('id', None)
            if rid is not None and rid in rid2step.keys():
                raise errors.DuplicatedId(rid)
            if o is None or str(o.id) != json_step.get('orchestration_id'):
                o = Orchestration.query.get_or_404(json_step.pop('orchestration_id'))
            else:
                json_step.pop('orchestration_id')
            if 'action_template_id' in json_step:
                json_step['action_template'] = ActionTemplate.query.get_or_404(json_step.pop('action_template_id'))
            elif 'action_type' in json_step:
                json_step['action_type'] = ActionType[json_step.pop('action_type')]
            dependencies[rid] = {'parent_step_ids': json_step.pop('parent_step_ids', []),
                                 'child_step_ids': json_step.pop('child_step_ids', [])}
            s = o.add_step(**json_step)
            db.session.add(s)
            new_id_steps.append(str(s.id))
            if rid:
                rid2step[rid] = s

            continue

        # process dependencies
        for rid, dep in dependencies.items():
            step = rid2step[rid]
            parents = []
            for p_s_id in dep['parent_step_ids']:
                if p_s_id in rid2step:
                    parents.append(rid2step[p_s_id])
                else:
                    parents.append(Step.query.get_or_404(p_s_id))
            o.set_parents(step, parents)
            children = []
            for c_s_id in dep['child_step_ids']:
                if c_s_id in rid2step:
                    children.append(rid2step[c_s_id])
                else:
                    children.append(Step.query.get_or_404(c_s_id))
            o.set_children(step, children)

        db.session.commit()

        return {'step_id': new_id_steps[0]} if isinstance(json_data, dict) else {'step_ids': new_id_steps}, 201


class StepResource(Resource):
    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self, step_id):
        return Step.query.get_or_404(step_id).to_json()

    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(step_put)
    @lock_catalog
    def put(self, step_id):
        s: Step = Step.query.get_or_404(step_id)
        json_data = request.get_json()
        s.undo = json_data.pop('undo')
        s.action_template = ActionTemplate.query.get_or_404(json_data.pop('action_template_id'))
        # remove dependencies
        s.orchestration.set_parents(s, [])
        s.orchestration.set_children(s, [])

        parent_step_ids = json_data.pop('parent_step_ids', [])
        parent_steps = []
        for parent_step_id in parent_step_ids:
            cs = Step.query.get_or_404(parent_step_id)
            parent_steps.append(cs)
            s.orchestration.set_parents(s, parent_steps)

        child_step_ids = json_data.pop('child_step_ids', [])
        child_steps = []
        for child_step_id in child_step_ids:
            cs = Step.query.get_or_404(child_step_id)
            child_steps.append(cs)
            s.orchestration.set_children(s, child_steps)

        s.stop_on_error = json_data.pop('stop_on_error', None)
        s.stop_undo_on_error = json_data.pop('stop_undo_on_error', None)
        s.undo_on_error = json_data.pop('undo_on_error', None)
        s.expected_stdout = json_data.pop('expected_stdout', None)
        s.expected_stderr = json_data.pop('expected_stderr', None)
        s.expected_rc = json_data.pop('expected_rc', None)
        s.parameters = json_data.pop('parameters', None)
        s.system_kwargs = json_data.pop('system_kwargs', None)
        s.target = json_data.pop('target', None)

        db.session.commit()

        return {}, 204

    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(step_patch)
    @lock_catalog
    def patch(self, step_id):
        s: Step = Step.query.get_or_404(step_id)
        json_data = request.get_json()
        if 'undo' in json_data:
            s.undo = json_data.pop('undo')
        if 'action_template_id' in json_data:
            s.action_template = ActionTemplate.query.get_or_404(json_data.pop('action_template_id'))
        if 'parent_step_ids' in json_data:
            parent_step_ids = json_data.pop('parent_step_ids')
            parent_steps = []
            for parent_step_id in parent_step_ids:
                ps = Step.query.get_or_404(parent_step_id)
                parent_steps.append(ps)
                s.orchestration.add_parents(s, parent_steps)
        if 'child_step_ids' in json_data:
            child_step_ids = json_data.pop('child_step_ids')
            child_steps = []
            for child_step_id in child_step_ids:
                cs = Step.query.get_or_404(child_step_id)
                child_steps.append(cs)
                s.orchestration.add_children(s, child_steps)

        for k, v in json_data.items():
            setattr(s, k, v)

        db.session.commit()

        return {}, 204

    @forward_or_dispatch
    @jwt_required
    @securizer
    @lock_catalog
    def delete(self, step_id):
        s: Step = Step.query.get_or_404(step_id)
        s.orchestration.delete_step(s)

        db.session.delete(s)
        db.session.commit()

        return {}, 204



import signal
import sys
from subprocess import *

if sys.version_info >= (3, 5):
    from subprocess import run as _run

    run = _run
else:

    # Exception classes used by this module.
    class SubprocessError(Exception):
        pass


    class CalledProcessError(SubprocessError):
        """Raised when run() is called with check=True and the process
        returns a non-zero exit status.

        Attributes:
          cmd, returncode, stdout, stderr, output
        """

        def __init__(self, returncode, cmd, output=None, stderr=None):
            self.returncode = returncode
            self.cmd = cmd
            self.output = output
            self.stderr = stderr

        def __str__(self):
            if self.returncode and self.returncode < 0:
                try:
                    return "Command '%s' died with %r." % (
                        self.cmd, signal.Signals(-self.returncode))
                except ValueError:
                    return "Command '%s' died with unknown signal %d." % (
                        self.cmd, -self.returncode)
            else:
                return "Command '%s' returned non-zero exit status %d." % (
                    self.cmd, self.returncode)

        @property
        def stdout(self):
            """Alias for output attribute, to match stderr"""
            return self.output

        @stdout.setter
        def stdout(self, value):
            # There's no obvious reason to set this, but allow it anyway so
            # .stdout is a transparent alias for .output
            self.output = value


    class TimeoutExpired(SubprocessError):
        """This exception is raised when the timeout expires while waiting for a
        child process.

        Attributes:
            cmd, output, stdout, stderr, timeout
        """

        def __init__(self, cmd, timeout, output=None, stderr=None):
            self.cmd = cmd
            self.timeout = timeout
            self.output = output
            self.stderr = stderr

        def __str__(self):
            return ("Command '%s' timed out after %s seconds" %
                    (self.cmd, self.timeout))

        @property
        def stdout(self):
            return self.output

        @stdout.setter
        def stdout(self, value):
            # There's no obvious reason to set this, but allow it anyway so
            # .stdout is a transparent alias for .output
            self.output = value


    class CompletedProcess(object):
        """A process that has finished running.

        This is returned by run().

        Attributes:
          args: The list or str args passed to run().
          returncode: The exit code of the process, negative for signals.
          stdout: The standard output (None if not captured).
          stderr: The standard error (None if not captured).
        """

        def __init__(self, args, returncode, stdout=None, stderr=None):
            self.args = args
            self.returncode = returncode
            self.stdout = stdout
            self.stderr = stderr

        def __repr__(self):
            args = ['args={!r}'.format(self.args),
                    'returncode={!r}'.format(self.returncode)]
            if self.stdout is not None:
                args.append('stdout={!r}'.format(self.stdout))
            if self.stderr is not None:
                args.append('stderr={!r}'.format(self.stderr))
            return "{}({})".format(type(self).__name__, ', '.join(args))

        def check_returncode(self):
            """Raise CalledProcessError if the exit code is non-zero."""
            if self.returncode:
                raise CalledProcessError(self.returncode, self.args, self.stdout,
                                         self.stderr)


    def run(*popenargs, input=None, timeout=None, check=False, **kwargs):
        """Run command with arguments and return a CompletedProcess instance.

        The returned instance will have attributes args, returncode, stdout and
        stderr. By default, stdout and stderr are not captured, and those attributes
        will be None. Pass stdout=PIPE and/or stderr=PIPE in order to capture them.

        If check is True and the exit code was non-zero, it raises a
        CalledProcessError. The CalledProcessError object will have the return code
        in the returncode attribute, and output & stderr attributes if those streams
        were captured.

        If timeout is given, and the process takes too long, a TimeoutExpired
        exception will be raised.

        There is an optional argument "input", allowing you to
        pass a string to the subprocess's stdin.  If you use this argument
        you may not also use the Popen constructor's "stdin" argument, as
        it will be used internally.

        The other arguments are the same as for the Popen constructor.

        If universal_newlines=True is passed, the "input" argument must be a
        string and stdout/stderr in the returned object will be strings rather than
        bytes.
        """
        if input is not None:
            if 'stdin' in kwargs:
                raise ValueError('stdin and input arguments may not both be used.')
            kwargs['stdin'] = PIPE

        with Popen(*popenargs, **kwargs) as process:
            try:
                stdout, stderr = process.communicate(input, timeout=timeout)
            except TimeoutExpired:
                process.kill()
                stdout, stderr = process.communicate()
                raise TimeoutExpired(process.args, timeout, output=stdout,
                                     stderr=stderr)
            except:
                process.kill()
                process.wait()
                raise
            retcode = process.poll()
            if check and retcode:
                raise CalledProcessError(retcode, process.args,
                                         output=stdout, stderr=stderr)
        return CompletedProcess(process.args, retcode, stdout, stderr)



import json
import os
from subprocess import CompletedProcess as CP, CompletedProcess
from unittest import mock

import responses
from aioresponses import aioresponses
from flask import url_for
from flask_jwt_extended import create_access_token
from pyfakefs.fake_filesystem_unittest import TestCase as FSTestCase

from dm import defaults
from dm.domain.entities import Server, Route, Dimension, User, ActionTemplate, ActionType, Orchestration, OrchExecution, \
    StepExecution
from dm.domain.entities.bootstrap import set_initial
from dm.network.auth import HTTPBearerAuth
from dm.use_cases.use_cases import upgrade_catalog
from dm.web import create_app, db
from tests.helpers import set_callbacks, generate_dimension_json_data


class TestLaunchOrchestrationNested(FSTestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.client = self.app.test_client()

        self.app2 = create_app('test')
        self.app2_context = self.app2.app_context()
        self.app_context.push()
        self.client2 = self.app2.test_client()

        self.json_dim = generate_dimension_json_data()

        set_initial(server=False, action_template=True)
        dim = Dimension.from_json(self.json_dim)
        db.session.add(dim)
        self.node1 = Server('node1', port=8000, me=True, id="00000000-1111-0000-0000-000000000000")
        db.session.add_all([self.node1])
        db.session.commit()
        self.auth = HTTPBearerAuth(create_access_token(User.get_by_user('root').id))

        # dump data
        self.json_node1 = self.node1.to_json(add_gates=True)

        with self.app2.app_context():
            set_initial(server=False, action_template=True)
            dim = Dimension.from_json(self.json_dim)
            db.session.add(dim)
            me = Server('node2', port=8000, me=True, granules=['backend'], id="00000000-2222-0000-0000-000000000000")
            db.session.add(me)

            src_server = Server.from_json(self.json_node1)
            Route(src_server, cost=0)
            db.session.add(src_server)

            db.session.commit()

            # dump data
            self.json_node2 = me.to_json(add_gates=True)

        self.node2 = Server.from_json(self.json_node2)
        r = Route(self.node2, cost=0)
        db.session.add_all([self.node2, r])
        db.session.commit()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

        with self.app2_context:
            db.session.remove()
            db.drop_all()

    def set_initial_data_nested_orchestration(self):
        # create data for orchestration
        at = ActionTemplate(name='mkdir', version=1, action_type=ActionType.SHELL, code='mkdir -f {{folder}}',
                            expected_rc=0)
        self.o = Orchestration(name='Create folder', version=1, description="creates a folder")
        self.s = self.o.add_step(undo=False, action_template=at)

        db.session.add_all([at, self.o, self.s])
        db.session.commit()
        self.o2 = Orchestration(name="launch create folder", version=1)

        self.s21 = self.o2.add_step(undo=False,
                                    action_type=ActionType.SHELL,
                                    code="ls",
                                    post_process="vc.set('hosts', {'all': '00000000-2222-0000-0000-000000000000'})"
                                    )

        self.s22 = self.o2.add_step(undo=False,
                                    action_template=ActionTemplate.query.filter_by(name='orchestration',
                                                                                   version=1).one(),
                                    parameters={'orchestration_id': str(self.o.id)},
                                    parents=[self.s21]
                                    )

        db.session.add_all([self.o2, self.s21, self.s22])

        db.session.commit()

        self.dtos = {'ActionTemplate': [at.to_json()],
                     'Orchestration': [self.o.to_json(), self.o2.to_json()],
                     'Step': [self.s.to_json(), self.s21.to_json(), self.s22.to_json()]}

        with self.app2.app_context():
            upgrade_catalog(self.dtos, check_mismatch=False)

    @mock.patch('dm.use_cases.operations.subprocess.run', autospec=True)
    @responses.activate
    @aioresponses()
    def test_launch_orchestration_nested_orchestration(self, mock_run, m):
        set_callbacks([(r'(127\.0\.0\.1|node1)', self.client), ('node2', self.client2)], m)

        mock_run.side_effect = [CP((), returncode=0, stdout='ls output', stderr=''),
                                CP((), returncode=0, stdout='folder output', stderr='')]

        ###### INITIAL DATA #####
        self.set_initial_data_nested_orchestration()

        resp = self.client.post(url_for('api_1_0.launch_orchestration', orchestration_id=self.o2.id),
                                json={'hosts': 'node1', 'params': {'folder': '/new_folder'}, 'background': False},
                                headers=self.auth.header)

        self.assertIn(resp.status_code, (200, 202))
        self.assertEqual(2, mock_run.call_count)
        self.assertTupleEqual(('ls',), mock_run.call_args_list[0][0])
        self.assertTupleEqual(('mkdir -f /new_folder',), mock_run.call_args_list[1][0])

        oe2: OrchExecution = OrchExecution.query.filter_by(orchestration_id=self.o2.id).one()
        self.assertEqual(User.get_by_user('root'), oe2.executor)
        self.assertEqual(True, oe2.success)
        self.assertDictEqual({'folder': '/new_folder'}, oe2.params)
        self.assertEqual(2, len(oe2.step_executions))
        se21: StepExecution = StepExecution.query.filter_by(step=self.s21).one()
        self.assertEqual(True, se21.success)
        self.assertEqual('ls output', se21.stdout)
        self.assertEqual('', se21.stderr)
        self.assertEqual(0, se21.rc)
        self.assertEqual(self.node1, se21.server)

        oe1: OrchExecution = OrchExecution.query.filter_by(orchestration_id=self.o.id).one()
        self.assertEqual(User.get_by_user('root'), oe1.executor)
        self.assertEqual(True, oe1.success)
        self.assertDictEqual({'folder': '/new_folder',
                              'hosts': {'all': "00000000-2222-0000-0000-000000000000"},
                              'orchestration_id': str(self.o.id),
                              'server_id': "00000000-1111-0000-0000-000000000000"}, oe1.params)
        self.assertEqual(1, len(oe1.step_executions))

        o_id = self.o.id
        with self.app2.app_context():
            oe1: OrchExecution = OrchExecution.query.filter_by(orchestration_id=o_id).one()
            self.assertEqual(1, len(oe1.step_executions))
            se1: StepExecution = oe1.step_executions[0]
            self.assertEqual(True, se1.success)
            self.assertEqual('folder output', se1.stdout)
            self.assertEqual('', se1.stderr)
            self.assertEqual(0, se1.rc)
            self.assertEqual(self.json_node2['id'], str(se1.server.id))

        se22: StepExecution = StepExecution.query.filter_by(step=self.s22).one()
        self.assertEqual(True, se22.success)
        self.assertIsNone(se22.stdout)

        self.assertEqual('', se21.stderr)
        self.assertEqual(0, se21.rc)
        self.assertEqual(self.node1, se22.server)

    def set_initial_data_install_software(self):
        self.source_folder = '/source'
        self.dest_folder = '/dest'
        self.filename = 'Python-3.8.3.tgz'
        self.content = '1234'
        self.size = 4
        self.checksum = "4"
        os.makedirs(self.source_folder)
        os.makedirs(self.dest_folder)
        with open(os.path.join(self.source_folder, self.filename), 'w') as fh:
            fh.write(self.content)

        soft_dto = [{"name": "python",
                     "version": "3.8.3",
                     "family": "programming",
                     "filename": self.filename,
                     "size": self.size,
                     "checksum": self.checksum,
                     "id": "00000019-0000-0000-0000-000000000001",
                     "last_modified_at": defaults.INITIAL_DATEMARK.strftime(defaults.DATEMARK_FORMAT)}]

        ssa_dto = [{'software_id': "00000019-0000-0000-0000-000000000001",
                    'server_id': "00000000-1111-0000-0000-000000000000",
                    'path': self.source_folder,
                    "last_modified_at": defaults.INITIAL_DATEMARK.strftime(defaults.DATEMARK_FORMAT)}]

        orch_dto = [{"name": "install python 3.8.3",
                     "version": 1,
                     "description": "installs python 3.8.3",
                     "id": "00000015-0000-0000-0001-000000000000",
                     "last_modified_at": defaults.INITIAL_DATEMARK.strftime(defaults.DATEMARK_FORMAT)
                     }]

        step_dtos = [
            {
                "id": "00000015-0000-0000-0001-000000000001",
                "orchestration_id": "00000015-0000-0000-0001-000000000000",
                "action_template_id": "00000000-0000-0000-000a-000000000001",
                "undo": False,
                "undo_on_error": False,
                "parameters": {
                    "software_id": "00000019-0000-0000-0000-000000000001",
                    "dest_path": "{{folder}}"},
                "last_modified_at": defaults.INITIAL_DATEMARK.strftime(defaults.DATEMARK_FORMAT)},
            {
                "id": "00000015-0000-0000-0001-000000000002",
                "orchestration_id": "00000015-0000-0000-0001-000000000000",
                "undo": False,
                "undo_on_error": False,
                "parent_step_ids": ["00000015-0000-0000-0001-000000000001"],
                "code": "echo '{{file}}'",
                "action_type": "SHELL",
                "expected_rc": 0,
                "last_modified_at": defaults.INITIAL_DATEMARK.strftime(defaults.DATEMARK_FORMAT)}]

        self.dtos = {'Software': soft_dto,
                     'SoftwareServerAssociation': ssa_dto,
                     'Orchestration': orch_dto,
                     'Step': step_dtos}
        upgrade_catalog(self.dtos, check_mismatch=False)

        with self.app2.app_context():
            upgrade_catalog(self.dtos, check_mismatch=False)

    @mock.patch('dm.use_cases.operations.subprocess.run')
    @responses.activate
    @aioresponses()
    def test_launch_orchestration_install_software_local(self, mock_run, m):
        set_callbacks([(r'(127\.0\.0\.1|node1)', self.client), ('node2', self.client2)], m)
        self.setUpPyfakefs()

        self.set_initial_data_install_software()

        mock_run.return_value = CompletedProcess(args=f"echo '{os.path.join(self.dest_folder, self.filename)}'",
                                                 stdout='output',
                                                 returncode=0)

        resp = self.client.post(
            url_for('api_1_0.launch_orchestration', orchestration_id="00000015-0000-0000-0001-000000000000"),
            json={'hosts': 'node1', 'params': {'folder': self.dest_folder}, 'background': False},
            headers=self.auth.header)

        self.assertEqual(200, resp.status_code)
        data = resp.get_json()
        self.assertTrue(data['success'])
        self.assertEqual(2, len(data['steps']))
        exec_step1 = data['steps'][0]
        exec_step2 = data['steps'][1]
        self.assertEqual(201, exec_step1.get('rc'))
        step_out = json.loads(exec_step1.get('stdout'))
        self.assertEqual(os.path.join(self.dest_folder, self.filename), step_out.get('file'))
        self.assertIn('id', step_out)
        self.assertDictEqual(
            dict(folder=self.dest_folder, dest_path=self.dest_folder,
                 file=os.path.join(self.dest_folder, self.filename), software_id="00000019-0000-0000-0000-000000000001",
                 server_id=self.node1.id),
            exec_step1.get('params'))

        self.assertEqual(0, exec_step2.get('rc'))
        self.assertEqual('output', exec_step2.get('stdout'))
        self.assertEqual(None, exec_step2.get('stderr'))
        self.assertDictEqual(
            dict(file=os.path.join(self.dest_folder, self.filename), folder=self.dest_folder, server_id=self.node1.id),
            exec_step2.get('params'))

        self.assertTupleEqual((f"echo '{os.path.join(self.dest_folder, self.filename)}'",), mock_run.call_args[0])
        self.assertTrue(os.path.exists(os.path.join(self.dest_folder, self.filename)))

    @mock.patch('dm.use_cases.operations.subprocess.run')
    @responses.activate
    @aioresponses()
    def test_launch_orchestration_install_software_remote(self, mock_run, m):
        set_callbacks([(r'(127\.0\.0\.1|node1)', self.client), ('node2', self.client2)], m)
        self.setUpPyfakefs()

        self.set_initial_data_install_software()

        mock_run.return_value = CompletedProcess(args=f"echo '{os.path.join(self.dest_folder, self.filename)}'",
                                                 stdout='output',
                                                 returncode=0)

        resp = self.client.post(
            url_for('api_1_0.launch_orchestration', orchestration_id="00000015-0000-0000-0001-000000000000"),
            json={'hosts': 'node2', 'params': {'folder': self.dest_folder}, 'background': False},
            headers=self.auth.header)

        self.assertEqual(200, resp.status_code)
        data = resp.get_json()
        self.assertTrue(data['success'])
        self.assertEqual(2, len(data['steps']))
        exec_step1 = data['steps'][0]
        exec_step2 = data['steps'][1]
        self.assertEqual(201, exec_step1.get('rc'))
        step_out = json.loads(exec_step1.get('stdout'))
        self.assertEqual(os.path.join(self.dest_folder, self.filename), step_out.get('file'))
        self.assertIn('id', step_out)
        self.assertDictEqual(
            dict(folder=self.dest_folder, dest_path=self.dest_folder,
                 file=os.path.join(self.dest_folder, self.filename), software_id="00000019-0000-0000-0000-000000000001",
                 server_id=self.node2.id),
            exec_step1.get('params'))

        self.assertEqual(0, exec_step2.get('rc'))
        self.assertEqual('output', exec_step2.get('stdout'))
        self.assertEqual(None, exec_step2.get('stderr'))
        self.assertDictEqual(
            dict(file=os.path.join(self.dest_folder, self.filename), folder=self.dest_folder, server_id=self.node2.id),
            exec_step2.get('params'))

        self.assertTupleEqual((f"echo '{os.path.join(self.dest_folder, self.filename)}'",), mock_run.call_args[0])
        self.assertTrue(os.path.exists(os.path.join(self.dest_folder, self.filename)))

    @mock.patch('dm.use_cases.operations.subprocess.run')
    @responses.activate
    @aioresponses()
    def test_launch_orchestration_install_software_dual(self, mock_run, m):
        set_callbacks([(r'(127\.0\.0\.1|node1)', self.client), ('node2', self.client2)], m)
        self.setUpPyfakefs()

        self.set_initial_data_install_software()

        mock_run.return_value = CompletedProcess(args=f"echo '{os.path.join(self.dest_folder, self.filename)}'",
                                                 stdout='output',
                                                 returncode=0)

        resp = self.client.post(
            url_for('api_1_0.launch_orchestration', orchestration_id="00000015-0000-0000-0001-000000000000"),
            json={'hosts': ['node1', 'node2'], 'params': {'folder': self.dest_folder}, 'background': False},
            headers=self.auth.header)

        self.assertEqual(200, resp.status_code)
        data = resp.get_json()
        self.assertTrue(data['success'])


from dm.domain.entities import Server, Locker, User, ActionTemplate
from dm.web import db


def set_initial(app=None, server=True, user=True, action_template=False):
    if app:
        app.app_context().push()
    db.create_all()
    Locker.set_initial()
    if server:
        Server.set_initial()
    if user:
        User.set_initial()
    if action_template:
        ActionTemplate.set_initial()
    if app:
        app.app_context().pop()



import datetime as dt
import os
from unittest import TestCase, mock

from dm.domain.entities import bypass_datamark_update, catalog, Orchestration, Catalog, Server, Route
from dm.web import db, create_app

basedir = os.path.abspath(os.path.dirname(__file__))


class TestInit(TestCase):

    def setUp(self) -> None:
        self.maxDiff = None
        self.app = create_app('test')
        self.app.config['SECURIZER'] = False
        # user sqlite in a file as flush does not emits states in an inmemory database
        # self.app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///'+os.path.join(basedir, 'test.db')
        self.app_context = self.app.app_context()
        self.app_context.push()
        db.create_all()

        # def remove_db():
        #     try:
        #         os.remove(os.path.join(basedir, 'test.db'))
        #     except FileNotFoundError:
        #         pass
        #
        # self.addCleanup(remove_db)

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_bypass_datamark_update(self):
        self.assertTrue(getattr(catalog, 'datemark', True))
        with bypass_datamark_update():
            self.assertFalse(catalog.datemark)
        self.assertTrue(catalog.datemark)

    @mock.patch('dm.domain.entities.get_now')
    def test_sqlalchemy_entity_events(self, mock_now):
        mock_now.return_value = dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc)

        o = Orchestration('orch', 1, last_modified_at=dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc))
        self.assertEqual(dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc), o.last_modified_at)
        db.session.add(o)
        self.assertEqual(dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc), o.last_modified_at)
        db.session.commit()
        self.assertEqual(dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc), o.last_modified_at)

        c = Catalog.query.get('Orchestration')
        self.assertEqual(dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc), c.last_modified_at)

        mock_now.return_value = dt.datetime(2019, 4, 3, tzinfo=dt.timezone.utc)

        o.name = 'modified'
        self.assertEqual(dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc), o.last_modified_at)
        db.session.flush()
        self.assertEqual(dt.datetime(2019, 4, 3, tzinfo=dt.timezone.utc), o.last_modified_at)
        db.session.commit()
        self.assertEqual(dt.datetime(2019, 4, 3, tzinfo=dt.timezone.utc), o.last_modified_at)

        c = Catalog.query.get('Orchestration')
        self.assertEqual(dt.datetime(2019, 4, 3, tzinfo=dt.timezone.utc), c.last_modified_at)

        mock_now.return_value = dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc)
        o2 = Orchestration('orch', 2, last_modified_at=dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc))
        db.session.add(o2)
        self.assertEqual(dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc), o2.last_modified_at)

        with bypass_datamark_update():
            self.assertEqual(dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc), o2.last_modified_at)

            o3 = Orchestration('orch', 3, last_modified_at=dt.datetime(2019, 4, 4, tzinfo=dt.timezone.utc))
            db.session.add(o3)
            o.name = 'orch'

            self.assertEqual(dt.datetime(2019, 4, 3, tzinfo=dt.timezone.utc), o.last_modified_at)
            self.assertEqual(dt.datetime(2019, 4, 4, tzinfo=dt.timezone.utc), o3.last_modified_at)

        self.assertEqual(dt.datetime(2019, 4, 3, tzinfo=dt.timezone.utc), o.last_modified_at)
        self.assertEqual(dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc), o2.last_modified_at)
        self.assertEqual(dt.datetime(2019, 4, 4, tzinfo=dt.timezone.utc), o3.last_modified_at)
        db.session.commit()
        self.assertEqual(dt.datetime(2019, 4, 3, tzinfo=dt.timezone.utc), o.last_modified_at)
        self.assertEqual(dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc), o2.last_modified_at)
        self.assertEqual(dt.datetime(2019, 4, 4, tzinfo=dt.timezone.utc), o3.last_modified_at)

        c = Catalog.query.get('Orchestration')
        self.assertEqual(dt.datetime(2019, 4, 4, tzinfo=dt.timezone.utc), c.last_modified_at)

    @mock.patch('dm.domain.entities.get_now')
    def test_sqlalchemy_entity_events_server(self, mock_now):
        mock_now.return_value = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)
        s = Server('source', port=5000)
        d = Server('destination', port=5000)
        db.session.add_all([s, d])
        db.session.commit()

        mock_now.return_value = dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc)

        s.route = Route(destination=d, gate=d.gates[0], cost=0)
        db.session.commit()
        self.assertEqual(dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc), s.last_modified_at)

        g = s.add_new_gate(dns_or_ip='dns', port=5000)
        db.session.add(g)
        db.session.commit()
        self.assertEqual(dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc), s.last_modified_at)



# -*- coding: utf-8 -*-

# pygtail - a python "port" of logtail2
# Copyright (C) 2011 Brad Greenlee <brad@footle.org>
#
# Derived from logcheck <http://logcheck.org>
# Copyright (C) 2003 Jonathan Middleton <jjm@ixtab.org.uk>
# Copyright (C) 2001 Paul Slootman <paul@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from __future__ import print_function

import glob
import gzip
import os
import sys
import typing as t
from os import fstat, stat
from os.path import exists, getsize

from dm.utils.mixins import LoggerMixin

PY3 = sys.version_info[0] == 3
MAX_LINE_SIZE = 250 * 1024 * 1024

if PY3:
    text_type = str
else:
    text_type = unicode


def force_text(s, encoding='utf-8', errors='strict'):
    if isinstance(s, text_type):
        return s
    return s.decode(encoding, errors)


class Pygtail(LoggerMixin):
    """
    Creates an iterable object that returns only unread lines.

    Parameters
    ----------
    offset_file:
       File to which offset data is written (default: <logfile>.offset).
    offset_mode:
       Sets when the offset is updated.
            'manual' updates the offset when set_offset called.
            'every' updates the offset file every n'th line. Parameter every_n must be given.
            'paranoid' updates the offset file every time we read a line.
            'endfile' updates the offset when we reach the end of the file.
    every_n:
        number of lines until the offset will be updated when 'every' offset mode set.
    copytruncate:
        Support copytruncate-style log rotation (default: True)
    on_update:
        Execute this function when offset data is written (default None)
    read_from_end:
        starts reading the file from the end discarding initial content
    log_patterns:
        List of custom rotated log patterns to match (default: None)
    full_lines:
        Only log when line ends in a newline (default: False)
    new_line:
        characters of new line. If None sets according to the OS
    binary:
        reads the file in binary mode.
    encoding:
        sets the encoding for the file
    errors:
        The errors argument specifies the response when the input string can’t be converted according to the
        encoding’s rules. See built-in open function for more details

    """

    def __init__(self, file, offset_file=None, offset_mode='endfile', copytruncate=True,
                 on_update: t.Callable = None, read_from_end=False, log_patterns=None, full_lines=False,
                 binary=False, encoding=None, errors=None, every_n=None):
        self.file = file
        if offset_mode in ('manual', 'every', 'paranoid', 'endfile'):
            if offset_mode == 'every':
                try:
                    self.every_n = int(every_n)
                except TypeError:
                    raise AttributeError("Parameter 'every_n' must be an integer when offset_mode in every")
            else:
                self.every_n = None
        else:
            raise AttributeError(f"invalid offset_mode '{offset_mode}'. See docstring for more details")
        self.offset_mode = offset_mode
        self.on_update = on_update
        self.copytruncate = copytruncate
        self.read_from_end = read_from_end
        self.log_patterns = log_patterns
        self.binary = binary
        self.encoding = encoding
        self.errors = errors
        self._full_lines = full_lines
        self._offset_file = offset_file or "%s.offset" % self.file
        self._offset_file_inode = 0
        self._offset = 0
        self._since_update = 0
        self._fh = None
        self._rotated_logfile = None

        # if offset file exists and non-empty, open and parse it
        if exists(self._offset_file) and getsize(self._offset_file):
            with open(self._offset_file, "r") as offset_fh:
                (self._offset_file_inode, self._offset) = [int(line.strip()) for line in offset_fh]
            if self._offset_file_inode != self.file_id or \
                    stat(self.file).st_size < self._offset:
                # The inode has changed or filesize has reduced so the file
                # might have been rotated.
                # Look for the rotated file and process that if we find it.
                self._rotated_logfile = self._determine_rotated_logfile()

    @property
    def file_id(self):
        data = stat(self.file)
        return data.st_ino or data.st_ctime_ns

    def __eq__(self, other):
        return isinstance(other, self.__class__) and other.file_id == self.file_id

    def __del__(self):
        if self._filehandle():
            self._filehandle().close()

    def __iter__(self):
        return self

    def next(self):
        """
        Return the next line in the file, updating the offset.
        """
        try:
            line = self._get_next_line()
        except StopIteration:
            # we've reached the end of the file; if we're processing the
            # rotated log file or the file has been renamed, we can continue with the actual file; otherwise
            # update the offset file
            if self._is_new_file():
                self._rotated_logfile = None
                self._fh.close()
                self._offset = 0
                # open up current logfile and continue
                try:
                    line = self._get_next_line()
                except StopIteration:  # oops, empty file
                    if self.offset_mode == 'endfile':
                        self._update_offset_file()
                        raise
            else:
                if self.offset_mode == 'endfile':
                    self._update_offset_file()
                raise

        if self.offset_mode == 'paranoid':
            self._update_offset_file()
        elif self.offset_mode == 'every' and self.every_n and self.every_n <= self._since_update:
            self._update_offset_file()

        return line

    def __next__(self):
        """`__next__` is the Python 3 version of `next`"""
        return self.next()

    def readline(self):
        try:
            return next(self)
        except StopIteration:
            return b'' if self.binary else ''

    def readlines(self, max_lines=None) -> t.List:
        """
        Read in all unread lines and return them as a list.
        """
        lines = []
        line = '\n'
        while (max_lines is None or len(lines) < max_lines) and len(line) != 0:
            line = self.readline()
            if line:
                lines.append(line)
        return lines

    def read(self, max_lines=None):
        """
        Read in all unread lines and return them as a single string.
        """
        lines = self.readlines(max_lines)
        if lines:
            # try:
            #     return ''.join(lines)
            # except TypeError:
            #     return ''.join(force_text(line) for line in lines)
            return b''.join(lines) if self.binary else ''.join(lines)
        else:
            return b'' if self.binary else ''

    def _is_closed(self):
        if not self._fh:
            return True
        try:
            return self._fh.closed
        except AttributeError:
            if isinstance(self._fh, gzip.GzipFile):
                # python 2.6
                return self._fh.fileobj is None
            else:
                raise

    def _filehandle(self):
        """
        Return a filehandle to the file being tailed, with the position set
        to the current offset.
        """
        if not self._fh or self._is_closed():
            file = self._rotated_logfile or self.file
            if file.endswith('.gz'):
                self._fh = gzip.open(file, 'rb' if self.binary else 'rt', encoding=self.encoding,
                                     errors=self.errors)
            else:
                self._fh = open(file, 'rb' if self.binary else 'r', 1, encoding=self.encoding, errors=self.errors)
            if self.read_from_end and not exists(self._offset_file):
                self._fh.seek(0, os.SEEK_END)
            else:
                self._fh.seek(self._offset)

        return self._fh

    def _update_offset_file(self):
        """
        Update the offset file with the current inode and offset.
        """
        if self.on_update:
            self.on_update()
        offset = self._filehandle().tell()
        fh = open(self._offset_file, "w")
        fh.write("%s\n%s\n" % (self.file_id, offset))
        fh.close()
        self._since_update = 0

    def update_offset_file(self):
        self._update_offset_file()

    def _determine_rotated_logfile(self):
        """
        We suspect the logfile has been rotated, so try to guess what the
        rotated file is, and return it.
        """
        rotated_filename = self._check_rotated_filename_candidates()
        if rotated_filename and exists(rotated_filename):
            data = stat(rotated_filename)
            if (data.st_ino or data.st_ctime_ns) == self._offset_file_inode:
                return rotated_filename

            # if the inode hasn't changed, then the file shrank; this is expected with copytruncate,
            # otherwise print a warning
            if self.file_id == self._offset_file_inode:
                if self.copytruncate:
                    return rotated_filename
                else:
                    sys.stderr.write(
                        "[pygtail] [WARN] file size of %s shrank, and copytruncate support is "
                        "disabled (expected at least %d bytes, was %d bytes).\n" %
                        (self.file, self._offset, stat(self.file).st_size))

        return None

    def _check_rotated_filename_candidates(self):
        """
        Check for various rotated logfile file patterns and return the first
        match we find.
        """
        # savelog(8)
        candidate = "%s.0" % self.file
        if (exists(candidate) and exists("%s.1.gz" % self.file) and
                (stat(candidate).st_mtime > stat("%s.1.gz" % self.file).st_mtime)):
            return candidate

        # logrotate(8)
        # with delaycompress
        candidate = "%s.1" % self.file
        if exists(candidate):
            return candidate

        # without delaycompress
        candidate = "%s.1.gz" % self.file
        if exists(candidate):
            return candidate

        rotated_filename_patterns = [
            # logrotate dateext rotation scheme - `dateformat -%Y%m%d` + with `delaycompress`
            "%s-[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9]",
            # logrotate dateext rotation scheme - `dateformat -%Y%m%d` + without `delaycompress`
            "%s-[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9].gz",
            # logrotate dateext rotation scheme - `dateformat -%Y%m%d-%s` + with `delaycompress`
            "%s-[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9]-[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9]",
            # logrotate dateext rotation scheme - `dateformat -%Y%m%d-%s` + without `delaycompress`
            "%s-[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9]-[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9].gz",
            # for TimedRotatingFileHandler
            "%s.[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]",
        ]
        if self.log_patterns:
            rotated_filename_patterns.extend(self.log_patterns)

        # break into directory and file components to support cases where
        # the file is prepended as part of rotation
        file_dir, rel_filename = os.path.split(self.file)
        for rotated_filename_pattern in rotated_filename_patterns:
            candidates = glob.glob(os.path.join(file_dir, rotated_filename_pattern % rel_filename))
            if candidates:
                candidates.sort()
                return candidates[-1]  # return most recent

        # no match
        return None

    def _is_new_file(self):
        # Processing rotated logfile or at the end of current file which has been renamed
        data = fstat(self._filehandle().fileno())
        return self._rotated_logfile or \
               self._filehandle().tell() == data.st_size and (data.st_ino or data.st_ctime_ns) != self.file_id

    def _get_next_line(self) -> t.Union[str, bytes]:

        curr_offset = self._filehandle().tell()
        line = self._filehandle().readline(MAX_LINE_SIZE)
        if self._full_lines:
            if self.binary and not line.endswith(self.new_line or os.linesep):
                self._filehandle().seek(curr_offset)
                if len(line) == MAX_LINE_SIZE:
                    self.logger.warning(f"Max Line Size reached while trying to get line from file '{self.file}' "
                                        f"at position {curr_offset}")
                raise StopIteration
            elif not line.endswith('\n'):
                self._filehandle().seek(curr_offset)
                if len(line) == MAX_LINE_SIZE:
                    self.logger.warning(f"Max Line Size reached while trying to get line from file '{self.file}' "
                                        f"at position {curr_offset}")
                raise StopIteration
        if not line:
            raise StopIteration
        self._since_update += 1
        return line



import argparse
import inspect
import shlex

from prompt_toolkit import PromptSession
from prompt_toolkit.history import InMemoryHistory

import dm.dshell.network as ntwrk
from dm.dshell.argparse_raise import create_parser, ArgumentParserRaise
from dm.dshell.commands import nested_dict
from dm.dshell.completer import DshellCompleter
from dm.dshell.helpers import exit_dshell, get_history


def call_func_with_signature(cmd_params):
    func = cmd_params.pop('func')
    sig = inspect.signature(func)
    args = []
    kwargs = {}
    arg_sig = False
    kw_sig = False

    for param in sig.parameters.values():
        if param.kind == param.POSITIONAL_ONLY or (
                param.kind == param.POSITIONAL_OR_KEYWORD and param.default == param.empty):
            args.append(cmd_params.pop(param.name))
        elif param.kind in (param.POSITIONAL_OR_KEYWORD, param.KEYWORD_ONLY):
            kwargs.update({param.name: cmd_params.pop(param.name, None)})
        elif param.kind == param.VAR_POSITIONAL:
            arg_sig = True
        elif param.kind == param.VAR_KEYWORD:
            kw_sig = True

    if arg_sig and not kw_sig:
        arg_sig = list(cmd_params.values())
        kw_sig = {}
    elif not arg_sig and kw_sig:
        arg_sig = []
        kw_sig = cmd_params
    else:
        arg_sig = []
        kw_sig = {}
    args = args + arg_sig
    func(*args, **{**kwargs, **kw_sig})

def interactive():
    session = PromptSession(completer=DshellCompleter.from_nested_dict(nested_dict),
                            history=get_history('main', InMemoryHistory()),
                            enable_history_search=True, enable_suspend=True)

    parser = ArgumentParserRaise(allow_abbrev=False, prog='')
    parser = create_parser(nested_dict, parser)
    while True:
        try:
            text = session.prompt(f'{ntwrk._username if ntwrk._username else "?"}.dshell> ')
        except KeyboardInterrupt:
            continue  # Control-C pressed. Try again.
        except EOFError:
            exit_dshell()
        try:
            namespace = parser.parse_args(shlex.split(text))
        except (ValueError, argparse.ArgumentError) as e:
            print(e)
            continue
        except SystemExit:
            continue
        else:
            cmd_params = dict(namespace._get_kwargs())

        if 'func' not in cmd_params:
            print(f"No action set for this command: {text}")
        else:
            call_func_with_signature(cmd_params)



"""
Usage: dshell action list [--id ID|--name NAME|--like LIKE] [--version N] [--last N]
       dshell action create (- | FILE)
       dshell action create NAME ACTION_TYPE [--code CODE] [--parameters DATA] [--expected-stdout OUT]
                            [--expected-stderr ERR] [--expected-rc RC] [--system-kwargs DATA]
                            [--pre-process CODE] [--post-process CODE]

Command to interact with action templates

Arguments:
    ACTION_TYPE     Type of action templates. Available: ANSIBLE, PYTHON, SHELL, ORCHESTRATION, REQUEST
    NAME            Action template name

Options:
    --code CODE             Code to execute
    --expected-rc RC        Expected return code from execution
    --expected-stderr ERR   Expected err output from execution
    --expected-stdout OUT   Expected output from execution
    --id ID                 Action template id to show
    --last N                Shows last N servers
    --like LIKE             List action templates whose names contain LIKE
    --name NAME             Action templates' name to list
    --parameters DATA       Parameters passed to the code
    --post-process POST     Code executed after CODE execution
    --pre-process PRE       Code executed before CODE execution
    --system-kwargs DATA    Special parameters passed to python command call
    --version N             Versions to list
"""
import json
import sys
from pprint import pprint

from docopt import docopt
from schema import Schema, Or, And, Use, SchemaError

import dm.dshell.network as ntwrk
from dm.dshell.commands import action_list
from dm.dshell.utils import clean_none

action_template_schema = Schema(
    {'ACTION_TYPE': Or(None, And(lambda x: x in ['ANSIBLE', 'PYTHON', 'SHELL', 'ORCHESTRATION', 'REQUEST'],
                                 error="Not a valid action type. See 'dshell help action' for more details")),
     'FILE': Or(None, Use(open, error="Unable to open file. File not found")),
     '--expected-rc': Or(None,
                         And(Use(int, error="Not a valid integer"),
                             And(lambda n: 0 <= n <= 255,
                                 error="integer must be between 0 and 255"))),
     '--last': Or(None, And(Use(int, error="Not a valid integer"), And(lambda n: 0 < n,
                                                                       error="integer must be above 0"))),
     str: object}, )


def main(args):
    argv = docopt(__doc__, args)

    try:
        argv = action_template_schema.validate(argv)
    except SchemaError as e:
        exit(str(e))
    if argv['list']:
        action_list(iden=argv['--id'], name=argv['--name'], version=argv['--version'], like=argv['--like'],
                    last=argv['--last'])
    elif argv['create']:
        if argv['-']:
            content = sys.stdin.read()
            part = json.loads(content)
            data = part if isinstance(part, list) else [part]
        elif argv['FILE']:
            data = []
            content = argv['FILE'].read()
            part = json.loads(content)
            data.extend(part) if isinstance(part, list) else data.append(part)
        else:
            data = {'name': argv['NAME'],
                    'version': argv['--version'],
                    'action_type': argv['ACTION_TYPE'],
                    'code': argv['--code'],
                    'parameters': argv['--parameters'],
                    'expected_stdout': argv['--expected-stdout'],
                    'expected_stderr': argv['--expected-stderr'],
                    'expected_rc': argv['--expected-rc'],
                    'system_kwargs': argv['--system-kwargs'],
                    'pre_process': argv['--pre-process'],
                    'post_process': argv['--post-process']
                    }
            data = clean_none(data)
        pprint(data)
        if data:
            resp = ntwrk.post(f'api_1_0.actiontemplatelist', json=data)
            if resp.ok:
                if resp.msg:
                    pprint(resp.msg)
            else:
                pprint(resp.msg if resp.code is not None else str(resp.exception) if str(
                    resp.exception) else resp.exception.__class__.__name__)
        else:
            print("No data found to create action template")



from flask import url_for
from flask_jwt_extended import create_access_token

from dm.domain.entities import Orchestration
from dm.domain.entities.bootstrap import set_initial
from dm.network.auth import HTTPBearerAuth
from dm.web import create_app, db
from tests.helpers import TestCaseLockBypass


class Test(TestCaseLockBypass):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.auth = HTTPBearerAuth(create_access_token('00000000-0000-0000-0000-000000000001'))
        set_initial()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_orchestration_resource_list(self):
        resp = self.client.get(url_for('api_1_0.orchestrationlist'), headers=self.auth.header)

        self.assertEqual([], resp.get_json())

        o = Orchestration(name='orchestration_name', version=1, description='desc')

        db.session.add(o)
        db.session.commit()

        resp = self.client.get(url_for('api_1_0.orchestrationlist'), headers=self.auth.header)

        self.assertEqual([o.to_json(add_target=True, add_params=True)], resp.get_json())

    def test_orchestration_resource(self):
        resp = self.client.post(url_for('api_1_0.orchestrationlist'),
                                json=dict(name='orchestration_name',
                                          version=1, description='desc'),
                                headers=self.auth.header)

        self.assertIn('orchestration_id', resp.get_json())

        o_id = resp.get_json()['orchestration_id']

        o = Orchestration.query.get(o_id)

        resp = self.client.get(url_for('api_1_0.orchestrationresource', orchestration_id=o_id),
                               headers=self.auth.header)
        db.session.refresh(o)
        self.assertDictEqual(o.to_json(add_target=True, add_params=True), resp.get_json())

        self.assertTrue(o.stop_on_error)
        self.assertTrue(o.stop_undo_on_error)
        resp = self.client.patch(url_for('api_1_0.orchestrationresource', orchestration_id=o_id),
                                 json={"stop_on_error": False},
                                 headers=self.auth.header)
        self.assertEqual(204, resp.status_code)
        db.session.refresh(o)
        self.assertFalse(o.stop_on_error)
        self.assertTrue(o.stop_undo_on_error)
